/***************************************************************************
 *                                                                         *
 *   ENTTLC.C                                                    Rev M     *
 *                                                                         *
 *   Copyright (C) 1987-1993 GALACTICOMM, Inc.    All Rights Reserved.     *
 *                                                                         *
 *   This is the Major BBS Entertainment Edition teleconference handler.   *
 *                                                                         *
 *                                            - T. Stryker 3/2/89          *
 *    Action Enhancements and Mods            - S. Brinker & R. Skurnick   *
 *                                            - 3/25/89                    *
 *    Tele-Games Enhancements and Mods        - J. Kobal 6/21/90           *
 *                                                                         *
 ***************************************************************************/

#include "gcomm.h"
#include "majorbbs.h"
#include "message.h"
#include "galetl.h"
#include "flash.h"
#include "bbsutils.h"

STATIC int tlclog(void);
STATIC int telecn(void);
STATIC void tcsthn(void);
STATIC void tlchup(void);
STATIC void deltlc(char *userid);
STATIC void clstlc(void);

struct module enttlcmd={      /* module interface block               */
     "",                      /*    name used to refer to this module */
     tlclog,                  /*    user logon supplemental routine   */
     telecn,                  /*    input routine if selected         */
     tcsthn,                  /*    status-input routine if selected  */
     NULL,                    /*    "injoth" routine for this module  */
     NULL,                    /*    user logoff supplemental routine  */
     tlchup,                  /*    hangup (lost carrier) routine     */
     NULL,                    /*    midnight cleanup routine          */
     deltlc,                  /*    delete-account routine            */
     clstlc                   /*    finish-up (sys shutdown) routine  */
};

static
int tlcstt;                   /* teleconferencing state number             */

#define NEEDCO -1             /* target user-id count when colon is needed */

FILE *tlcmb;                  /* teleconf named-message file block pointer */
static
BTVFILE *tbb;                 /* BTRIEVE file pointer for tele. profiles   */

struct tlc *tlclst,           /* one to a customer (dynamic array)         */
           *tlcptr,*tptr;     /* handy pointers for speed                  */

struct tlc2 *tl2lst,          /* one to a customer (dynamic array)         */
           *tl2ptr;           /* handy pointers for speed                  */

#define CHNOFF 10000          /* channel offset for private channels       */

#define FNDBIT(u) (1L<<((u)&0x1F))
#define INVFLS(u) invites[(u)/32]
#define FORFLS(u) forgets[(u)/32]

static
struct tlcdat {               /* teleconference profile data structure     */
     char userid[UIDSIZ];     /*   user-id (BTRIEVE key)                   */
     int defchn;              /*   default teleconference channel          */
     int action;              /*   default action status (on/off)          */
     int pagsts;              /*   default page status (on/off/ok)         */
     int pagint;              /*   default page interval (1-9 mins)        */
     int chaint;              /*   default chat interval (1-9 mins)        */
     char theme[TPCSIZ];      /*   default private channel topic           */
     char entmsg[TMSSIZ];     /*   customized entrance message             */
     char extmsg[TMSSIZ];     /*   customized exit message                 */
     int flags;               /*   various flags that need to be saved     */
     char spare[252-203];     /*   spare space (decrease when adding)      */
} tlcdat;                     /* occurence of structure for lookups        */

                              /* definition of data file user bit "flags"  */
#define TSYSOP  1             /*   user is a Teleconference Sysop          */
#define CANGA   2             /*   user has Generic Action capability      */

static
struct telemisc {             /* miscellaneous teleconference data         */
     int edtin;               /*   sysop editor is in use                  */
     int wchmsg;              /*   which message is being edited           */
     int entchg;              /*   entrance message changed boolean        */
     int extchg;              /*   exit message changed boolean            */
     int updrec;              /*   record must be updated                  */
     char uname[UIDSIZ];      /*   who's being changed (user-id)           */
} *telmisc;                   /* dynamic allocated occurence of structure  */

                              /* bit flag definitions for updrec           */
#define UPDFLG 1              /*   updated one of the users flags          */
#define UPDMSG 2              /*   updated one of the users messages       */

/*--- OPTIONS FROM GALETL.MSG ---*/

int gbldbt;                   /* allow users to gamble when in debt?       */

static
int starsq,                   /* want asterisks in front of async msgs?    */
    npaymx,                   /* max times can talk per session if no pay  */
    nswchx,                   /* max times can switch chans or on-off/15sec*/
    tinpsz,                   /* allowable input size for teleconference   */
    sopbel,                   /* Sysop page at main console bell period    */
    dftpop,                   /* default page option setting (ON/OFF/OK)   */
    oktocg,                   /* ok to charge for teleconference messages  */
    dispad,                   /* display charges on auditrail              */
    lnkaud,                   /* generate audit trail entry per linkup?    */
    msgchg,                   /* amount of charges if ok to charge         */
    etlccr,                   /* teleconference credit consumption rate    */
    etlcov,                   /* does the tele cc rate over-ride Forums'?  */
    eswtfr,                   /* allow usrs to swt to & from Forum chans?  */
    etlpst;                   /* number of credits a TSYSOP can post       */


static
char *linkpw,                 /* incoming link-up request password         */
     *diahay,                 /* outgoing dial prefix for Hayes-category   */
     *diaxec,                 /* outgoing dial prefix for XECOM-category   */
     *gacstg,                 /* string for initiating a generic action    */
     *somoth;                 /* displaying names of users in telecon      */

STATIC int ck4pfn(void);
STATIC void g2chan(int chno);
STATIC int undupe(char *stg,int insert);
STATIC void linkup(void);
STATIC void lvchi(int eximsg,int ntrmsg);
STATIC void dial(void);
STATIC int chnavl(int chan);
STATIC char tlchat(int chan,int ch);
STATIC int quitlc(void);
STATIC void outbt2(unsigned chn);
STATIC void tlcctx(void);
STATIC void tpfmsg();
STATIC void tpfmlt();
STATIC int findstar(char *stg);
STATIC int glopag(void);
STATIC void page(void);
STATIC void pospag(char *pagmsg);
STATIC void chat(void);
STATIC void invite(void);
STATIC void devite(void);
STATIC void unifor(void);
STATIC void forget(void);
STATIC void rememb(void);
STATIC void join(void);
STATIC int chncnt(void);
STATIC char *thisher(int usr);
STATIC void whisper(char *whoto,int mnum);
STATIC char *prspow(char *start,int mnum,int prec,int *count);
STATIC int howmny(char *stg,int prec,int samas);
STATIC void tlctck(void);
STATIC void initls(void);
STATIC void outfgm(void);
STATIC char *tlsrui(void);
STATIC void ret2tl(int who);
STATIC void hupchi(int xitmsg);
STATIC void actas(int usrnm);
STATIC void echo(void);
STATIC void entrmsg(void);
STATIC int urinv(void);
STATIC void action(void);
STATIC int chkact(void);
STATIC void simples(void);
STATIC void dobject(void);
STATIC void bothsnd(void);
STATIC void yelling(void);
STATIC void generic(void);
STATIC int actacc(void);
STATIC void inidya(void);
STATIC struct action *actbin(char *stgptr);
STATIC void teditor(void);
STATIC void tedtmnu(void);
STATIC void edtr01(void);
STATIC void edtr02(void);
STATIC void edtr03(void);
STATIC void edtr04(void);
STATIC void edtr05(void);
STATIC void edtr06(void);
STATIC void usredt(void);
STATIC void chkhdl(void);
STATIC void pmtedw(void);
STATIC void edtwch(void);
STATIC void updflag(int flag);
STATIC void shomsg(void);
STATIC void updmsg(void);
STATIC void dedcds(void);
STATIC void chgmsg(void);

void EXPORT
init__enttlc(void)                 /* initialize teleconference module     */
{
     char *cp;

     stzcpy(enttlcmd.descrp,gmdnam("GALETL.MDF"),MNMSIZ);
     tlcstt=register_module(&enttlcmd);
     tlcmb=opnmsg("galetl.mcv");
     tbb=opnbtv("galetl.dat",sizeof(struct tlcdat));
     tlclst=(struct tlc *)alczer(nterms*sizeof(struct tlc));
     tl2lst=(struct tlc2 *)alczer(nterms*sizeof(struct tlc2));
     telmisc=(struct telemisc *)alcmem(sizeof(struct telemisc));
     setmem(telmisc,sizeof(struct telemisc),0);
     starsq=ynopt(STARSQ);
     gbldbt=ynopt(GBLDBT);
     npaymx=numopt(NPAYMX,0,32767);
     nswchx=numopt(NSWCHX,1,32767);
     tinpsz=numopt(TINPSZ,1,255);
     sopbel=numopt(SOPBEL,0,2000);
     cp=lastwd(rawmsg(DFTPOP));
     dftpop=*(cp+strlen(cp)-1);
     oktocg=ynopt(OKTOCG);
     dispad=ynopt(DISPAD);
     msgchg=numopt(MSGCHG,1,32767);
     linkpw=stgopt(LINKPW);
     lnkaud=ynopt(LNKAUD);
     diahay=stgopt(DIAHAY);
     diaxec=stgopt(DIAXEC);
     gacstg=stgopt(GACSTG);
     etlccr=numopt(ETLCCR,-32767,32767);
     etlcov=ynopt(ETLCOV);
     etlpst=numopt(ETLPST,0,9999);
     somoth=stgopt(SOMOTH);
     eswtfr=(etlcov || ynopt(ESWTFR));
     inidya();
     if (tjoinrou == NULL) {
          tjoinrou=joint;
     }
     rtkick(10,tlctck);
     globalcmd(glopag);
     initng();
     initpok();
     initbj();
     inifla();
}

void
joint(chan)                        /* join teleconference from other module*/
unsigned chan;
{
     tlcptr=&tlclst[usrnum];
     if (!(usrptr->flags&X2MAIN)) {
          tlcptr->flags|=RSTX2M;
          usrptr->flags|=X2MAIN;
     }
     tlcptr->retstt=usrptr->state;
     tlcptr->retsub=usrptr->substt;
     tlcptr->retrat=usrptr->crdrat;
     tlcptr->flags|=XTOOTH;
     tlcptr->channel=chan+1;
     usrptr->state=tlcstt;
     usrptr->substt=0;
     telecn();
}

STATIC int
tlclog(void)                       /* teleconference supplemental logon   */
{
     setbtv(tbb);
     setmbk(tlcmb);
     tlcptr=&tlclst[usrnum];
     tl2ptr=&tl2lst[usrnum];
     tlcptr->prvchn=usrnum+CHNOFF;
     if (!acqbtv(&tlcdat,usaptr->userid,0)) {
          setmem(&tlcdat,sizeof(struct tlcdat),0);
          strcpy(tlcdat.userid,usaptr->userid);
          switch (dftpop) {
          case 'N':
               tlcdat.pagsts=1;
               break;
          case 'F':
               tlcdat.pagsts=2;
               break;
          case 'K':
               tlcdat.pagsts=3;
               break;
          }
          tlcdat.defchn=1;
          tlcdat.action=1;
          tlcdat.pagint=8;
          tlcdat.chaint=8;
          insbtv(&tlcdat);
     }
     tlcptr->pflags=tlcdat.flags;
     if ((tlcptr->defchn=tlcdat.defchn) == 2) {
          if (hasmkey(TOPKEY)) {
               tlcptr->channel=usrnum+CHNOFF;
          }
     }
     if ((tlcptr->action=tlcdat.action) == 1) {
          if (hasmkey(ACTKEY)) {
               tlcptr->flags|=ACTION;
          }
     }
     switch ((tlcptr->pagsts=tlcdat.pagsts)) {
     case 1:  /* on */
          break;
     case 2:  /* off */
          tlcptr->flags|=NOPAGE;
          break;
     case 3:  /* ok  */
          tlcptr->flags|=OKPAGE;
          break;
     }
     tlcptr->pagint=tlcdat.pagint;
     tlcptr->chaint=tlcdat.chaint;
     strcpy(tl2ptr->entmsg,tlcdat.entmsg);
     strcpy(tl2ptr->extmsg,tlcdat.extmsg);
     if (hasmkey(TOPKEY)) {
          strcpy(tlcptr->theme,tlcdat.theme);
          strcpy(tlcptr->topic,tlcdat.theme);
     }
     if (usrptr->flags&MASTER) {
          tlcptr->pflags|=(TSYSOP+CANGA);
     }
     return(0);
}

STATIC int
telecn(void)                       /* main teleconferencing input handler  */
{
     int i,count;
     void (*rouptr)();
     struct tlc *tp;
     char uid[UIDSIZ],*ptr;

     setmbk(tlcmb);
     tlcptr=&tlclst[usrnum];
     tl2ptr=&tl2lst[usrnum];
     switch (usrptr->substt) {
     case 0:
          usrptr->substt=TLKING;
          btumil(usrnum,tinpsz);
          btuxnf(usrnum,0,19);
          if (!(usrptr->flags&INVISB)) {
               entrmsg();
               outtlc();
          }
          tpfmsg(INTRO);
          tlcctx();
          if (etlcov || !(tlcptr->flags&XTOOTH)) {
               usrptr->crdrat=etlccr;
          }
          break;
     case TLKING:
          if (margc == 0) {
               tlcctx();
               break;
          }
          if (input[inplen-1] == '^') {
               prfmsg(MSGABT);
               break;
          }
          if (!urinv()) {
               if (ck4fla()) {
                    break;
               }
          }
          else {
               clrprf();
          }
          if (margc == 1) {
               if (sameas(margv[0],"#") || sameas(margv[0],"users")) {
                    usrson();
                    prfmsg(ETLPMT);
                    outprf(usrnum);
                    clrprf();
                    break;
               }
               if (sameas(margv[0],"?") || sameas(margv[0],"/?")) {
                    tpfmsg(TLCHLP3);
                    break;
               }
               if (sameas(margv[0],"x") || sameas(margv[0],"/x")) {
                    if ((tlcptr->flags&JUSTEX) && !(usrptr->flags&OPCHAT)) {
                         tpfmsg(NANNOY);
                         break;
                    }
                    telext();
                    if (!(usrptr->flags&INVISB)) {
                         tlcptr->flags|=JUSTEX;
                         if (!strlen(tl2ptr->extmsg)) {
                              tpfmlt(LVITLC,usaptr->userid);
                         }
                         else {
                              tpfmlt(EXISKL,tl2ptr->extmsg);
                         }
                         outtlc();
                    }
                    tpfmsg(EXITLC);
                    return(quitlc());
               }
               if (sameas(margv[0],"menu") && hasmkey(MNUKEY)) {
                    if (telmisc->edtin) {
                         tpfmsg(EDTINU);
                         break;
                    }
                    telext();
                    if (!(usrptr->flags&INVISB)) {
                         tlcptr->flags|=JUSTEX;
                         tpfmlt(ENTEDT,usaptr->userid);
                         outtlc();
                    }
                    usredt();
                    break;
               }
               if (sameas(margv[0],"edit") || sameas(margv[0],"/e")) {
                    if ((tlcptr->flags&JUSTEX) && !(usrptr->flags&OPCHAT)) {
                         tpfmsg(NANNOY);
                         break;
                    }
                    telext();
                    if (!(usrptr->flags&INVISB)) {
                         tlcptr->flags|=JUSTEX;
                         tpfmlt(ENTEDT,usaptr->userid);
                         outtlc();
                    }
                    tedtmnu();
                    break;
               }
               if (sameas(margv[0],"main") || sameas(margv[0],"/m")) {
                    if (tlcptr->channel == 0) {
                         tpfmsg(ALRMAN);
                    }
                    else if (!eswtfr && tlcptr->channel < CHNOFF
                             && !(usrptr->flags&MASTER)) {
                         tpfmsg(CANTEX);
                    }
                    else {
                         g2chan(0);
                    }
                    break;
               }
               if (sameas(margv[0],"scan") || sameas(margv[0],"/s")) {
                    tpfmsg(NSCNHDR);
                    for (tp=tlclst,othusn=0 ; othusn < nterms ; tp++,othusn++) {
                         othusp=&user[othusn];
                         if (othusp->state == tlcstt && !(othusp->flags&INVISB)) {
                              prf("%-30s... ",uacoff(othusn)->userid);
                              if (tlcptr->FORFLS(othusn)&FNDBIT(othusn)) {
                                   prf("* ");
                              }
                              else {
                                   prf("  ");
                              }
                              if (tlcptr->INVFLS(othusn)&FNDBIT(othusn)) {
                                   prf("* ... ");
                              }
                              else {
                                   prf("  ... ");
                              }
                              if (othusp->substt == CHTING) {
                                   prf(" (Chatting)");
                              }
                              else if (othusp->substt == LNKING) {
                                   prf(" (Linking)");
                              }
                              else if (othusp->substt >= EDTING) {
                                   prf(" (Editing)");
                              }
                              else if (tp->channel == 0) {
                                   prf("Main");
                              }
                              else if (tp->channel < CHNOFF) {
                                   prf("%s",sigoff(tp->channel-1)->signam);
                              }
                              else {
                                   prf("%s's",
                                       uacoff(tp->channel-CHNOFF)->userid);
                              }
                              if (tp->topic[0] != '\0'
                                && othusp->substt == TLKING
                                && tp->channel == tp->prvchn) {
                                   prf("\r%42s(%s)\r","",tp->topic);
                              }
                              else if (othusp->substt == FLHING) {
                                   pbptr=fapyrs[othusn].poolpt;
                                   prf(" [Flash %s %s]\r",pbptr->gamnam,
                                                          pbptr->gamver);
                              }
                              else {
                                   prf("\r");
                              }
                         }
                    }
                    break;
               }
               if (sameas(margv[0],"tingo")) {
                    tnghlp();
                    outfgm();
                    break;
               }
               if (sameas(margv[0],"poker")) {
                    pokhlp();
                    outfgm();
                    break;
               }
               if (sameas(margv[0],"bj")) {
                    bjhlp();
                    outfgm();
                    break;
               }
          }
          if (margc == 2) {
               if (sameas(margv[0],"action") || sameas(margv[0],"/a")) {
                    action();
                    break;
               }
               if (sameas(margv[0],"echo") || sameas(margv[0],"/h")) {
                    echo();
                    break;
               }
               if (sameas(margv[0],"tingo")) {
                    if (sameas(margv[1],"off")) {
                         tngoff();
                    }
                    else if (sameas(margv[1],"on")) {
                         tngon();
                    }
                    else if (sameas(margv[1],"card")) {
                         tshwcrd();
                    }
                    else if (sameas(margv[1],"scan")) {
                         tngscan();
                    }
                    else if (sameas(margv[1],"pot")) {
                         tngpot();
                    }
                    else {
                         tnghlp();
                    }
                    outfgm();
                    break;
               }
               if (sameas(margv[0],"invite") || sameas(margv[0],"/i")) {
                    invite();
                    break;
               }
               if (sameas(margv[0],"uninvite") || sameas(margv[0],"/u")) {
                    devite();
                    break;
               }
               if (sameas(margv[0],"forget") || sameas(margv[0],"/f")) {
                    forget();
                    break;
               }
               if (sameas(margv[0],"remember")) {
                    rememb();
                    break;
               }
          }
          if (margc >= 2 && sameas(margv[0],"poker")) {
               if (sameas(margv[1],"off")) {
                    pokoff();
               }
               else if (sameas(margv[1],"on")) {
                    pokon();
               }
               else if (sameas(margv[1],"hand")) {
                    pshwhnd();
               }
               else if (sameas(margv[1],"scan")) {
                    pokscan();
               }
               else if (sameas(margv[1],"turn")) {
                    pokturn();
               }
               else if (sameas(margv[1],"fold")) {
                    pokfold();
               }
               else if (sameas(margv[1],"pot")) {
                    pokpot();
               }
               else if (sameas(margv[1],"call")) {
                    pokcall();
               }
               else if (sameas(margv[1],"bet")) {
                    pokbet();
               }
               else if (sameas(margv[1],"raise")) {
                    pokrais();
               }
               else if (sameas(margv[1],"discard")) {
                    pokdscrd();
               }
               else if (sameas(margv[1],"low")) {
                    poklow();
               }
               else if (sameas(margv[1],"high")) {
                    pokhigh();
               }
               else {
                    pokhlp();
               }
               outfgm();
               break;
          }
          if (margc >= 2 && sameas(margv[0],"bj")) {
               if (sameas(margv[1],"off")) {
                    bjoff();
               }
               else if (sameas(margv[1],"on")) {
                    bjon();
               }
               else if (sameas(margv[1],"hand")) {
                    bshwhnd();
               }
               else if (sameas(margv[1],"scan")) {
                    bjscan();
               }
               else if (sameas(margv[1],"turn")) {
                    bjturn();
               }
               else if (sameas(margv[1],"fold")) {
                    bjfold();
               }
               else if (sameas(margv[1],"pot")) {
                    bjpot();
               }
               else if (sameas(margv[1],"call")) {
                    bjcall();
               }
               else if (sameas(margv[1],"bet")) {
                    bjbet();
               }
               else if (sameas(margv[1],"raise")) {
                    bjrais();
               }
               else if (sameas(margv[1],"hit")) {
                    bjhit();
               }
               else if (sameas(margv[1],"stay")) {
                    bjstay();
               }
               else if (sameas(margv[1],"low")) {
                    bjlow();
               }
               else if (sameas(margv[1],"high")) {
                    bjhigh();
               }
               else {
                    bjhlp();
               }
               outfgm();
               break;
          }
          if (sameas(margv[0],"chat") || sameas(margv[0],"/c")) {
               chat();
               break;
          }
          if (sameas(margv[0],"join") || sameas(margv[0],"/j")) {
               join();
               break;
          }
          if (hasmkey(DIALKY) && sameas(margv[0],"dial")) {
               telext();
               dial();
               break;
          }
          if (margc == 3 && sameas(margv[0],"link")) {
               telext();
               if (!sameas(margv[1],linkpw)) {
                    tpfmsg(LPWNOG);
                    break;
               }
               if (lnkaud) {
                    shocst("TELECONFERENCE LINK-UP","%s linked with %.7s",
                           usaptr->userid,margv[2]);
               }
               if (!(usrptr->flags&INVISB)) {
                    tpfmlt(XFORMD,usaptr->userid,margv[2]);
                    outtlc();
               }
               for (i=1 ; i < nmods ; i++) {
                    if (i != tlcstt && (rouptr=module[i]->huprou) != NULL) {
                         (*rouptr)();
                    }
               }
               setmbk(tlcmb);
               tlcptr=&tlclst[usrnum];
               tl2ptr=&tl2lst[usrnum];
               usaptr->usedat=today();
               updacc();
               usrptr->class=BBSPRV;
               usrptr->flags|=NOZAP;
               sprintf(usaptr->userid,"(%.7s)",margv[2]);
               btuech(usrnum,0);
               linkup();
               prfmsg(LINKDM);
               break;
          }
          if (margc >= 2 && sameas(margv[0],"post")
            && (tlcptr->pflags&TSYSOP)) {
               if (!etlpst) {
                    prfmsg(NOTPST);
               }
               else {
                    rstrin();
                    stzcpy(uid,margv[1],UIDSIZ);
                    zonkhl(uid);
                    switch (addcrd(uid,spr("%d",etlpst),0)) {
                    case -1:
                         prfmsg(BDPUID);
                         break;
                    case 1:
                         saycrd(spr("%d",etlpst),0);
                    case 0:
                         prfmsg(OKPUID,etlpst,uid);
                         break;
                    }
               }
               break;
          }
          if (margc >= 2 && (sameas(margv[0],"zappo") || sameas(margv[0],"/z"))
           && (tlcptr->pflags&TSYSOP)) {
               rstrin();
               if (margv[margc-1][0] == '/') {
                    margn[margc-2][0]='\0';
               }
               if (onsysn(margv[1],1)) {
                    i=usrnum;
                    curusr(othusn);
                    movmem(usaptr->userid,uid,UIDSIZ);
                    loscar();
                    curusr(i);
                    setmbk(tlcmb);
                    clrprf();
                    prfmsg(ZAPDOK);
                    if (margv[margc-1][0] == '/' && hasmkey(DELKEY)) {
                         if (tolower(margv[margc-1][1]) == 'd') {
                              if (delacct(uid) >= 0) {
                                   prfmsg(ZAPDDL);
                                   shocst("TELE-USER DELETED","%s deleted by %s"
                                          ,uid,usaptr->userid);
                              }
                              else {
                                   prfmsg(ZAPDND);
                              }
                         }
                         else if (tolower(margv[margc-1][1]) == 's') {
                              kiluid(uid);
                              prfmsg(ZAPDSU);
                         }
                    }
               }
               else {
                    prfmsg(ZAPDNO);
               }
               break;
          }
          if (margc >= 2) {
               if (sameas(margv[0],"invite") || sameas(margv[0],"/i")) {
                    invite();
                    break;
               }
               if (sameas(margv[0],"uninvite") || sameas(margv[0],"/u")) {
                    devite();
                    break;
               }
               if (sameas(margv[0],"forget") || sameas(margv[0],"/f")) {
                    forget();
                    break;
               }
               if (sameas(margv[0],"remember") || sameas(margv[0],"/r")) {
                    rememb();
                    break;
               }
          }
          if (sameas(margv[0],"page") || sameas(margv[0],"/p")) {
               page();
               break;
          }
          if (ck4pfn()) {
               break;
          }
          if (sameas(margv[0],"topic") || sameas(margv[0],"/t")) {
               if (urinv()) {
               }
               else if (margc == 1) {
                    tlcptr->topic[0]='\0';
                    tpfmsg(CLRTHE);
               }
               else if (!hasmkey(TOPKEY)) {
                    tlcptr->topic[0]='\0';
                    tpfmsg(NOTHEM);
                    howbuy();
               }
               else {
                    rstrin();
                    *(margv[1]+TPCSIZ-1)='\0';
                    strcpy(tlcptr->topic,margv[1]);
                    if (tlcptr->channel == tlcptr->prvchn) {
                         tpfmlt(NOWTHE,usaptr->userid,tlcptr->topic);
                         outtlc();
                    }
                    tpfmsg(URTHEM,tlcptr->topic);
               }
               break;
          }
          if (!hasmkey(UNLKEY) && (tlcptr->inpcnt)++ >= npaymx) {
               tpfmsg(NPAYXC,npaymx);
               howbuy();
               break;
          }
          if (sameas(margv[0],"whisper")) {
               if (urinv()) {
               }
               else if (margc < 4) {
                    tpfmsg(WHSFMT);
               }
               else {
                    whisper(margv[2],2);
               }
               break;
          }
          if (margv[0][0] == '/') {
               if (urinv()) {
               }
               else if (margc < 2 || margv[0][1] == '\0') {
                    tpfmsg(WHSFMT);
               }
               else {
                    whisper(margv[0]+1,0);
               }
               break;
          }
          if (chkact()) {
               break;
          }
          rstrin();
          tpfmlt(TLCFRM,usaptr->userid,input);
          outtlc();
          if (chncnt() == 1) {
               tpfmsg(BYSELF);
          }
          else {
               prfmsg(NMSENT);
          }
          break;
     case CHTING:
          if (margc == 1 && sameas(margv[0],"x")) {
               lvchi(EXICH2,ENTTLC);
               if (usrptr->flags&OPCHAT) {
                    clrprf();
                    actas(tlcptr->chatch);
                    lvchi(EXICH2,ENTTLC);
                    outprf(usrnum);
                    actas(tlcptr->chatch);
               }
          }
          else {
               return(1);
          }
          break;
     case LNKING:
          if (margc == 1 && (sameas(margv[0],"x") || sameas(margv[0],"xx"))) {
               if (usaptr->userid[0] == '(') {
                    rstchn();
                    return(1);
               }
               else {
                    lvchi(EXILNK,ENTTLC);
               }
          }
          else if (margc > 0 && (sameas(margn[margc-1]-7,"LINKED!")
                              || sameas(margn[margc-1]-14,"LLIINNKKEEDD!!"))) {
               if (usaptr->userid[0] != '(') {
                    lvchi(LINKOK,LKDTLC);
                    actas(tlcptr->chatch);
                    linkup();
                    actas(tlcptr->chatch);
               }
               else {
                    return(1);
               }
          }
          else if (usaptr->userid[0] == '(') {
               if (bdspec()) {
                    shochl("Linking Teleconferences",'',
                            baudat(usrptr->baud,0));
               }
               return(1);
          }
          else {
               return(1);
          }
          break;
     case LINKED:
          if (margc == 0) {
               return(1);
          }
          if (*margv[0] == ':') {
               margv[0]++;
          }
          if (margc == 1 && sameas(margv[0],"***")) {
               return(1);
          }
          for (i=0,count=0 ; i < margc ; i++) {
               if (*margv[i] != '(' && ++count >= 2) {
                    break;
               }
               if (sameas(margv[i],usaptr->userid)) {
                    return(1);
               }
          }
          rstrin();
          ptr=getmsg(NMSENT);
          stpans(ptr);
          if (!sameto(margv[i],ptr)
            && undupe(margv[0],0)
            && undupe(margv[0]+1,0)
            && undupe(margv[0]+2,0)) {
               tpfmlt(LNMSKL,usaptr->userid,margv[0]);
               outtlc();
          }
          return(1);
     case GOING:
     case FLHING:
          return(1);
     default:
          if (pfnlvl >= 2 && usrptr->pfnacc > WRNPFN) {
               tpfmsg(RAUNCH);
               break;
          }
          if (pfnlvl > 2) {
               tpfmsg(PFNWRD);
               break;
          }
          switch (usrptr->substt) {
          case 10:
               teditor();
               break;
          case 11:
               edtr01();
               break;
          case 12:
               edtr02();
               break;
          case 13:
               edtr03();
               break;
          case 14:
               edtr04();
               break;
          case 15:
               edtr05();
               break;
          case 16:
               edtr06();
               break;
          case 90:
               chkhdl();
               break;
          case 91:
               edtwch();
               break;
          case 92:
               updmsg();
               break;
          default:
               catastro("ENTTLC SUBSTT=%d ERROR",usrptr->substt);
          }
     }
     if (usrptr->substt == TLKING && *prfbuf != '\0') {
          prfmsg(ETLPMT);
     }
     outprf(usrnum);
     return(1);
}

STATIC int
ck4pfn(void)                  /* check for profanity and deal with it */
{
     int retval=0;

     if (usrptr->pfnacc > MAXPFN) {
          btuinj(usrnum,RING);
          prf("");
          retval=1;
     }
     else if (pfnlvl >= 2 && usrptr->pfnacc > WRNPFN) {
          prfmsg(RAUNCH);
          retval=1;
     }
     else if (pfnlvl > 2) {
          prfmsg(PFNWRD);
          retval=1;
     }
     return(retval);
}

STATIC void
g2chan(chno)                       /* go to specified teleconf channel #   */
{
     if (!(usrptr->flags&INVISB)) {
          tpfmlt(LEFTCH,usaptr->userid);
          outtlc();
     }
     tlcptr->channel=chno;
     if (!(usrptr->flags&INVISB)) {
          tpfmlt(CAMEIN,usaptr->userid);
          outtlc();
     }
     tlcctx();
}

STATIC int
undupe(stg,insert)                 /* utility for ignoring echoed output   */
char *stg;
int insert;
{
     int i;
     long thissg;
     static long signas[10];

     for (thissg=0L ; *stg != '\0' && *stg != '\r' ; stg++) {
          if (*stg == ESC) {
               do {
                    stg++;
               } while (*stg != '\0' && *stg != '\r' && !isalpha(*stg));
               if (!isalpha(*stg) || *++stg == '\0' || *stg == '\r') {
                    break;
               }
          }
          if ((thissg+=thissg) < 0L) {
               thissg=-thissg+1;
          }
          thissg+=*stg;
     }
     for (i=0 ; i < sizeof(signas)/sizeof(long) ; i++) {
          if (thissg == signas[i]) {
               return(0);
          }
     }
     if (insert) {
          movmem(signas,signas+1,sizeof(signas)-sizeof(long));
          signas[0]=thissg;
     }
     return(1);
}

STATIC void
linkup(void)                       /* enter linked-up condition            */
{
     btumil(usrnum,INPSIZ-1);
     btutsw(usrnum,0);
     shochl(spr("Linked with %s",usaptr->userid),'',baudat(usrptr->baud,0));
     btuxnf(usrnum,0,0);
     btuchi(usrnum,NULL);
     btucmd(usrnum,"]");
     usrptr->substt=LINKED;
}

STATIC void
lvchi(eximsg,ntrmsg)               /* leave the btuchi() "chat" condition  */
int eximsg,ntrmsg;
{
     tlcptr->flags&=~CCHVLD;
     tlcptr->reqcha=0;
     btuchi(usrnum,NULL);
     btuchi(tlcptr->chatch,NULL);
     btucli(usrnum);
     btumil(usrnum,tinpsz);
     tpfmlt(ntrmsg,usaptr->userid,uacoff(tlcptr->chatch)->userid);
     outtlc();
     usrptr->substt=TLKING;
     usrptr->flags&=~NOGLOB;
     tpfmsg(eximsg);
     tlcctx();
     echon();
}

STATIC void
dial(void)                         /* dial out to link to another system   */
{
     int saveun;
     char *ptr;
     struct tlc *otlptr;

     if (margc < 3) {
          tpfmsg(LNKSYN);
          return;
     }
     if (*margv[1] != '*') {
          othusn=0xFF;
          sscanf(margv[1],"%x",&othusn);
          if ((othusn=usridx(othusn)) < 0) {
               tpfmsg(NSUCHC,margv[1]);
               return;
          }
          if (!chnavl(othusn)) {
               tpfmsg(CHNINU,margv[1]);
               return;
          }
     }
     else {
          for (othusn=nterms-1 ; othusn >= 0 ; othusn--) {
               if (chnavl(othusn)) {
                    break;
               }
          }
          if (othusn < 0) {
               tpfmsg(NOCHAV);
               return;
          }
     }
     othusp=&user[othusn];
     othuap=uacoff(othusn);
     otlptr=&tlclst[othusn];
     if (!(usrptr->flags&INVISB)) {
          tpfmlt(ODIALG,usaptr->userid,margv[2]);
          outtlc();
     }
     if (othusp->flags&ISRIAL) {
          tpfmsg(SERIGO);
     }
     else if (margc < 4) {
          tpfmsg(NONUMD);
          return;
     }
     else if (othusp->flags&IS2698) {
          ptr=spr("%s%s\r",diahay,margv[3]);
          btuxct(othusn,strlen(ptr),ptr);
          tpfmsg(DIALOMP);
     }
     else {
          btucmd(othusn,spr("%s%sM",diaxec,margv[3]));
          tpfmsg(DIALOMP);
     }
     sprintf(othuap->userid,"(%.7s)",margv[2]);
     btuech(usrnum,0);
     btuech(othusn,0);
     tlcptr->chatch=othusn;
     otlptr->chatch=usrnum;
     btuchi(usrnum,tlchat);
     btuchi(othusn,tlchat);
     usrptr->substt=LNKING;
     othusp->substt=LNKING;
     othusp->state=tlcstt;
     othusp->class=BBSPRV;
     othusp->flags|=NOZAP;
     otlptr->channel=tlcptr->channel;
     saveun=usrnum;
     usrnum=othusn;
     shochl("Teleconference Dialing Out",'',baudat(usrptr->baud,0));
     usrnum=saveun;
}

STATIC int
chnavl(chan)                       /* channel is available for dial out?   */
int chan;
{
     struct user *othusp;

     othusp=&user[chan];
     return(othusp->class == VACANT
       && othusp->state == AWAITC
       && !(othusp->flags&NOHDWE));
}

STATIC char
tlchat(chan,ch)                    /* bypass routine for btuchi() for chat */
int chan;
int ch;
{
     int oth;
     char c;

     oth=tlclst[chan].chatch;
     c=((char)ch)&eurmsk;
     switch (c) {
     case '\r':
          chiinp(oth,c);
          chiout(oth,c);
          chiout(oth,'\n');
          return(c);
     case 8:
     case 127:
          chiinp(oth,8);
          chiout(oth,8);
          chiout(oth,' ');
          chiout(oth,8);
          return(8);
     default:
          if (c >= 32) {
               if (!(tlclst[chan].flags&TYPING)) {
                    if (uacoff(oth)->ansifl&ANSON) {
                         chious(oth,"\33[32m");
                    }
                    if (uacoff(chan)->ansifl&ANSON) {
                         chious(chan,"\33[33m");
                    }
                    tlclst[chan].flags|=TYPING;
                    tlclst[oth].flags&=~TYPING;
               }
               chiinp(oth,c);
               chiout(oth,c);
               return(c);
          }
     }
     return(0);
}

STATIC int
quitlc(void)                       /* quit teleconference, whence came     */
{
     rstrxf();
     if (tlcptr->flags&XTOOTH) {
          if (tlcptr->flags&RSTX2M) {
               usrptr->flags&=~X2MAIN;
          }
          tlcptr->flags&=~(XTOOTH+RSTX2M);
          tlcptr->channel=0;
          if (tlcptr->defchn == 2) {
               if (hasmkey(TOPKEY)) {
                    tlcptr->channel=usrnum+CHNOFF;
               }
          }
          condex();
          usrptr->state=tlcptr->retstt;
          usrptr->substt=tlcptr->retsub;
          usrptr->crdrat=tlcptr->retrat;
          injacr();
          return(1);
     }
     return(0);
}

STATIC void
outbt2(chn)                        /* prf() to all but usrnum and chn      */
unsigned chn;
{
     char sav,*ptr;

     ptr=prfptr;
     prfmlt(ETLPMT);
     sav=*ptr;
     for (othusn=0,othusp=user ; othusn < nterms ; othusn++,othusp++) {
          if (othusn != usrnum && othusp->state == tlcstt
            && othusp->substt == TLKING && othusn != chn) {
               if (tlclst[usrnum].channel == tlclst[othusn].channel) {
                    if (!(tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
                         switch (othusp->substt) {
                         case TLKING:
                              outmlt(othusn);
                              break;
                         case LINKED:
                              *ptr='\0';
                              undupe(prfbuf+findstar(prfbuf),1);
                              *ptr=sav;
                              *(ptr-1)='\0';
                              btuxmt(othusn,prfbuf+findstar(prfbuf));
                              btuxct(othusn,1,"\r");
                              *(ptr-1)='\r';
                              break;
                         }
                    }
               }
          }
     }
     clrmlt();
}

STATIC void
tlcctx(void)                       /* teleconference channel user display  */
{
     char *curguy,tmpbuf[UIDSIZ+6];
     unsigned chan;
     int othuno;

     initls();
     tlcptr=&tlclst[usrnum];
     chan=tlcptr->channel;
     othuno=(tlcptr->channel-CHNOFF);
     if (chan != 0) {
          if (chan < CHNOFF) {
               tpfmsg(SIGCHN,sigoff(chan-1)->signam);
          }
          else {
               if (chan == tlcptr->prvchn) {
                    tpfmsg(URTPRV);
               }
               else {
                    tpfmsg(HISPRV,uacoff(othuno)->userid);
               }
               if (tlclst[othuno].topic[0] != '\0') {
                    tpfmsg(PRVTOP,tlclst[othuno].topic);
               }
          }
     }
     else {
          tpfmsg(MAINTC);
     }
     switch (chncnt()) {
     case 0:
     case 1:
          tpfmsg(BYSELF);
          break;
     case 2:
          tpfmsg(ONEOTH,tlsrui());
          break;
     case 3:
          strcpy(tmpbuf,tlsrui());
          tpfmsg(TWOOTH,tmpbuf,tlsrui());
          break;
     default:
          strcpy(tmpbuf,tlsrui());
          while ((curguy=tlsrui()) != NULL) {
               prf(somoth,tmpbuf);
               strcpy(tmpbuf,curguy);
          }
          tpfmsg(SEVOTH,tmpbuf);
     }
     tpfmsg(IROEPI);
}

STATIC void
tpfmsg(msgno,parm1,parm2,parm3)    /* teleconf prfmsg-equiv, with star-strip */
int msgno;
long parm1,parm2,parm3;
{
     int cnt;
     char *tptr;

     tptr=getmsg(msgno);
     if (!starsq && (cnt=findstar(tptr)) != 0) {
          strcpy(tptr+cnt-4,tptr+cnt);
     }
     prf(tptr,parm1,parm2,parm3);
}

STATIC void
tpfmlt(msgno,parm1,parm2,parm3)    /* teleconf prfmlt-equiv, with star-strip */
int msgno;                              /* (butchered to meet requirements)  */
long parm1,parm2,parm3;
{
     char *tptr;
     int clsave,cnt;

     cklonl();
     clsave=clingo;
     for (clingo=0 ; clingo < nlingo ; clingo++) {
          if (linuse[clingo]) {
               prfptr=prfpointers[clingo];
               tptr=getmsg(msgno);
               if (!starsq && (cnt=findstar(tptr)) != 0) {
                    strcpy(tptr+cnt-4,tptr+cnt);
               }
               prf(tptr,parm1,parm2,parm3);
               prfpointers[clingo]=prfptr;
          }
     }
     prfptr=prfpointers[0];
     clingo=clsave;
     mltflg=1;
}

STATIC int
findstar(stg)                      /* find offset of end of stars (or 0)   */
char *stg;                              /* stg to find offset in           */
{
     int rv;
     char buf[40];

     stzcpy(buf,stg,sizeof(buf));
     stpans(buf);
     if (!(rv=(sameto("***\r",buf) ? findstg("***\r",stg) : 0))) {
          rv=(sameto("***\n",buf) ? findstg("***\n",stg) : 0);
     }
     return(rv);
}

STATIC int
glopag(void)                       /* globally-accessible form of "page"   */
{
     if (margc >= 1 && sameas(margv[0],"/p")) {
          setmbk(tlcmb);
          if (margc == 1 || (margc == 2 && sameas(margv[1],"?"))) {
               prfmsg(GLPFMT);
          }
          else {
               tlcptr=&tlclst[usrnum];
               page();
          }
          outprf(usrnum);
          return(1);
     }
     else if (margc == 1 && sameas(margv[0],"/#")) {
          usrson();
          outprf(usrnum);
          return(1);
     }
     return(0);
}

STATIC void
page(void)                         /* "page" function                      */
{
     int count;
     char *pagmsg;

     if (urinv()) {
     }
     else if (margc == 1) {
          tpfmsg(PAGFMT);
     }
     else if (*margv[1] == '(') {
          tpfmsg(NOPLNK);
     }
     else {
          rstrin();
          if ((count=howmny(margv[1],0,1)) == 1) {
               pagmsg="";
          }
          else {
               pagmsg=prspow(margv[1],1,0,&count);
          }
          if (count == 0) {
               if (sameas(margv[1],"on")) {
                    tlcptr->flags&=~(NOPAGE+OKPAGE);
                    tpfmsg(PAGTON,tlcptr->pagint/4,tlcptr->chaint/4);
               }
               else if (sameas(margv[1],"off")) {
                    tlcptr->flags|=NOPAGE;
                    tpfmsg(PAGTOF);
               }
               else if (sameas(margv[1],"ok")) {
                    tlcptr->flags|=OKPAGE;
                    tlcptr->flags&=~NOPAGE;
                    tpfmsg(PAGTOK);
               }
               else if (sameas(margv[1],"Sysop") || sameto("Sysop ",margv[1])) {
                    if (tlcptr->blinkc) {
                         tpfmsg(PAGS2M);
                    }
                    else {
                         belper(sopbel);
                         printf("\7");
                         tpfmsg(PAGEOK,"Sysop (at main console)");
                         psmatm();
                         shochl(usaptr->userid,'',baudat(usrptr->baud,1));
                         tlcptr->blinkc=8;
                    }
               }
               else {
                    tpfmsg(NPGUID,margv[1]);
               }
          }
          else if (count == NEEDCO) {
               prfmsg(PAMBIG,margv[1],"/p");
          }
          else if (count != 1) {
               tpfmsg(NPGAMB,margv[1]);
          }
          else {
               pospag(pagmsg);
          }
     }
}

STATIC void
pospag(pagmsg)                     /* possible page utility           */
char *pagmsg;
{
     if (!hasmkey(PAGKEY)) {
          tpfmsg(CANTPG);
          howbuy();
     }
     else if (tlclst[othusn].flags&NOPAGE) {
          tpfmsg(PAGOFF,othuap->userid);
     }
     else if ((tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
          tpfmsg(IGNORN,othuap->userid);
     }
     else if (!(tlclst[othusn].flags&OKPAGE) && tlclst[othusn].paged) {
          tpfmsg(PAGL2M,othuap->userid,tlclst[othusn].pagint/4);
     }
     else if (!ck4pfn()) {
          if (*pagmsg == '\0') {
               prfmlt(PAGMSG2,usaptr->userid,module[usrptr->state]->descrp);
          }
          else {
               prfmlt(PAGNOT2,usaptr->userid,
                              module[usrptr->state]->descrp,pagmsg);
          }
          if (injoth()) {
               tlclst[othusn].paged=tlclst[othusn].pagint;
               tpfmsg(PAGEOK,othuap->userid);
          }
          else {
               tpfmsg(PAGNPS,othuap->userid);
          }
     }
}

STATIC void
chat(void)                         /* "chat" function                      */
{
     int i,count;

     if (urinv()) {
     }
     else if (margc == 1) {
          tpfmsg(CHAFMT);
     }
     else {
          rstrin();
          if ((count=howmny(margv[1],0,1)) == 0
            || (count == 1 && othusp->substt >= EDTING)
            || (count == 1 && othusp->state != tlcstt)) {
               tpfmsg(NCHUID,margv[1]);
          }
          else if (count != 1) {
               tpfmsg(NCHAMB,margv[1]);
          }
          else if (othusp->substt == CHTING) {
               tpfmsg(ALRICH,othuap->userid);
          }
          else if (othusn == usrnum) {
               tpfmsg(NCHSLF);
          }
          else if ((tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
               tpfmsg(IGNORN,othuap->userid);
          }
          else if ((tlclst[othusn].flags&CCHVLD)
            && usrnum == tlclst[othusn].chatch) {
               tlclst[usrnum].chatch=othusn;
               tlclst[usrnum].flags&=~TYPING;
               tlclst[othusn].flags&=~TYPING;
               btumil(othusn,-(othuap->scnwid-1));
               btumil(usrnum,-(usaptr->scnwid-1));
               usrptr->substt=CHTING;
               othusp->substt=CHTING;
               usrptr->flags|=NOGLOB;
               othusp->flags|=NOGLOB;
               telext();
               tlcptr=&tlclst[othusn];
               i=usrnum;
               usrnum=othusn;
               telext();
               tpfmlt(GONCHA,othuap->userid);
               outtlc();
               btucli(othusn);
               btuclo(othusn);
               tpfmlt(ACCPCH,usaptr->userid);
               tpfmlt(ENTCHA);
               outmlt(othusn);
               btuchi(othusn,tlchat);
               usrnum=i;
               btuclo(usrnum);
               btucli(usrnum);
               tlcptr=&tlclst[usrnum];
               tpfmlt(GONCHA,usaptr->userid);
               outtlc();
               tpfmsg(ENTCHA);
               btuchi(usrnum,tlchat);
          }
          else if (!hasmkey(CHTKEY)) {
               tpfmsg(CANTCH);
               howbuy();
          }
          else if (tlclst[othusn].flags&NOPAGE) {
               tpfmsg(PAGOFF,othuap->userid);
          }
          else if (!(tlclst[othusn].flags&OKPAGE) && tlclst[othusn].reqcha) {
               tpfmsg(CHAL2M,othuap->userid,tlclst[othusn].chaint/4);
          }
          else {
               tpfmlt(CHAREQ,usaptr->userid,(usaptr->sex == 'M' ? "him" : "her"),
                    usaptr->userid);
               if (injoth()) {
                    tlcptr->chatch=othusn;
                    tlcptr->flags|=CCHVLD;
                    tlclst[othusn].reqcha=tlclst[othusn].chaint;
                    tpfmsg(CREQOK,othuap->userid);
               }
               else {
                    tpfmsg(PAGNPS,othuap->userid);
               }
          }
     }
}

STATIC void
invite(void)                       /* invite a user to a channel           */
{
     int count;
     struct tlc *otlcpt;

     if (urinv()) {
          return;
     }
     rstrin();
     count=howmny(margv[1],0,1);
     if (count == 0) {
          if (sameas(margv[1],"all")) {
               tlcptr->flags|=INVALL;
               tpfmsg(OKALLI);
          }
          else {
               tpfmsg(PAGNON,margv[1]);
          }
     }
     else if (count != 1) {
          tpfmsg(AMBIG,margv[1],"invite");
     }
     else {
          if (usrnum == othusn) {
               tpfmsg(WHYYOU,"invite");
          }
          else if ((tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
               tpfmsg(IGNORN,othuap->userid);
          }
          else if (othuap->userid[0] == '(') {
               tpfmsg(NOINVL);
          }
          else {
               tlcptr->INVFLS(othusn)|=FNDBIT(othusn);
               otlcpt=&tlclst[othusn];
               if (otlcpt->flags&NOPAGE
                 || ((otlcpt->paged) && !(otlcpt->flags&OKPAGE))) {
               }
               else {
                    tpfmlt(INVITA,usaptr->userid,
                                  (usaptr->sex == 'M' ? "his" : "her"));
                    if (othusp->state == tlcstt && othusp->substt == TLKING) {
                         tpfmlt(INVITJ,usaptr->userid);
                    }
                    injoth();
               }
               tpfmsg(OKINVT,othuap->userid);
               if (tlcptr->prvchn != tlcptr->channel) {
                    tpfmsg(INVWRN);
               }
          }
     }
}

STATIC void
devite(void)                       /* uninvite a user from a channel       */
{
     int count;

     if (urinv()) {
          return;
     }
     rstrin();
     count=howmny(margv[1],0,1);
     if (count == 0) {
          if (sameas(margv[1],"all")) {
               tlcptr->flags&=~INVALL;
               tpfmsg(UNVALL);
          }
          else {
               tpfmsg(PAGNON,margv[1]);
          }
     }
     else if (count != 1) {
          tpfmsg(AMBIG,margv[1],"uninvite");
     }
     else if (usrnum == othusn) {
          tpfmsg(WHYYOU,"uninvite");
     }
     else if (othusp->flags&MASTER) {
          tpfmsg(NODICE);
     }
     else {
          unifor();
     }
}

STATIC void
unifor(void)                       /* kicked out by forget or uninvite cmd */
{
     int i,j;
     struct tlc *tmptlc,*otlcpt;
     struct usracc *tmpptr;

     otlcpt=&tlclst[othusn];
     tlcptr->INVFLS(othusn)&=~FNDBIT(othusn);
     if (otlcpt->channel == tlcptr->prvchn) {
          otlcpt->channel=0;
          if (othusp->state == tlcstt && othusp->substt == FLHING) {
               ret2tl(othusn);
               tpfmsg(KUNVIT,othuap->userid);
          }
          else if (othusp->state == tlcstt && othusp->substt == TLKING) {

               tmptlc=tlcptr;
               tlcptr=otlcpt;
               j=usrnum;
               i=usrnum=othusn;

               tpfmlt(KICKDO,uacoff(i)->userid);
               prfmlt(ETLPMT);
               for (othusn=0,othusp=user ; othusn < nterms ; othusn++,othusp++) {
                    if (othusn != j && othusp->state == tlcstt
                      && othusp->substt == TLKING && othusn != i) {
                         if (tmptlc->channel == tlclst[othusn].channel) {
                              if (!(tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
                                   outmlt(othusn);
                              }
                         }
                    }
               }
               clrmlt();

               tmpptr=othuap;
               curusr(i);
               tpfmsg(UNVITD,usaptr->userid);
               tlcctx();
               prfmsg(ETLPMT);
               outprf(usrnum);

               tpfmlt(CAMEIN,uacoff(i)->userid);
               outtlc();
               tlcptr=tmptlc;
               curusr(j);

               tpfmsg(KUNVIT,uacoff(i)->userid);
               othuap=tmpptr;
          }
          else {
               tpfmsg(OKUNVT,othuap->userid);
          }
     }
     else {
          tpfmsg(OKUNVT,othuap->userid);
     }
}

STATIC void
forget(void)                       /* forgetting a user and uninvite       */
{
     int count;

     if (urinv()) {
          return;
     }
     rstrin();
     count=howmny(margv[1],0,1);
     if (count == 0) {
          tpfmsg(PAGNON,margv[1]);
     }
     else if (count != 1) {
          tpfmsg(AMBIG,margv[1],"forget");
     }
     else if (othusp->flags&MASTER) {
          tpfmsg(NODICE);
     }
     else {
          tlcptr->FORFLS(othusn)|=FNDBIT(othusn);
          unifor();
          clrprf();
          tpfmsg(FORGT2,othuap->userid,
                       (othuap->sex == 'M' ? "his" : "her"),
                        othuap->userid);
     }
}

STATIC void
rememb(void)                       /* remembering a user                   */
{
     int count;

     if (urinv()) {
          return;
     }
     rstrin();
     count=howmny(margv[1],0,1);
     if (count == 0) {
          tpfmsg(PAGNON,margv[1]);
     }
     else if (count != 1) {
          tpfmsg(AMBIG,margv[1],"remember");
     }
     else {
          tlcptr->FORFLS(othusn)&=~FNDBIT(othusn);
          tpfmsg(REMEMD,othuap->userid,(othuap->sex == 'F' ? "her" : "his"));
     }
}

STATIC void
join(void)                         /* joining a channel                    */
{
     int ochn,signo,count;

     if ((tlcptr->swchan)++ >= nswchx) {
          tpfmsg(NANNOY);
     }
     else if (tlcptr->channel != 0 && !eswtfr && tlcptr->channel < CHNOFF
              && !(usrptr->flags&MASTER)) {
          tpfmsg(CANTEX);
     }
     else if (margc == 1) {
          g2chan(tlcptr->channel == tlcptr->prvchn ? 0 : tlcptr->prvchn);
     }
     else if (*(nxtcmd=margv[1]) == SIGIDC) {
          qscptr=qscoff(usrnum);
          iniqsc();
          if (!eswtfr && !(usrptr->flags&MASTER)) {
               tpfmsg(CANTJN);
          }
          else if ((signo=findsig(cncuid())) == NOSIG) {
               tpfmsg(NOSAXS);
          }
          else if (tlcptr->channel == signo+1) {
               tpfmsg(ALRINS);
          }
          else {
               g2chan(signo+1);
          }
     }
     else {
          rstrin();
          count=howmny(margv[1],0,1);
          if (count == 0) {
               tpfmsg(PAGNON,margv[1]);
          }
          else if (count != 1) {
               tpfmsg(AMBIG,margv[1],"join");
          }
          else if (CHNOFF+othusn == tlcptr->channel) {
               tpfmsg(ALRTHR,(othuap->sex == 'F' ? "her" : "his"));
          }
          else if ((tlclst[othusn].INVFLS(usrnum)&FNDBIT(usrnum))
            || (tlclst[othusn].flags&INVALL)
            || (usrptr->flags&MASTER)) {
               ochn=othusn;
               g2chan(ochn+CHNOFF);
          }
          else {
               tpfmsg(NOTINV,othuap->userid);
          }
     }
}

STATIC int
chncnt(void)                       /* count number of users on channel     */
{
     int i,cnt;
     unsigned chan;

     chan=tlclst[usrnum].channel;
     for (i=0,cnt=0 ; i < nterms ; i++) {
          if (user[i].state == tlcstt
            && (user[i].substt == TLKING || user[i].substt == LINKED)
            && tlclst[i].channel == chan) {
               if (!(user[i].flags&INVISB) || i == usrnum) {
                    cnt++;
               }
          }
     }
     return(cnt);
}

STATIC char *
thisher(usr)                       /* returns "his" or "her" for a user    */
int usr;
{
     return(uacoff(usr)->sex == 'F' ? "her" : "his");
}

STATIC void
whisper(whoto,mnum)                /* "whisper" function                   */
char *whoto;
int mnum;
{
     int count;
     char *what;

     what=prspow(whoto,mnum,2,&count);
     if (count == NEEDCO) {
          prfmsg(PAMBIG,whoto,"whisper to");
     }
     else if (count == 0) {
          tpfmsg(WHSNHR,whoto);
     }
     else if (count != 1) {
          tpfmsg(AMBIG,whoto,"whisper");
     }
     else if (*what == '\0') {
          tpfmsg(WHSFMT);
     }
     else if ((tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum))) {
          tpfmsg(IGNORN,othuap->userid);
     }
     else {
          tpfmlt(WHSTO,usaptr->userid,what);
          prfmlt(ETLPMT);
          outmlt(othusn);
          prfmsg(WHSSNT,othuap->userid);
     }
}

STATIC char *
prspow(start,mnum,prec,count)      /* parse destination of some action     */
char *start;
int mnum,prec,*count;
{
     char *msg;
     int gobl,i;

     for (i=0 ; i < margc ; i++) {
          margn[i][0]='\0';
     }
     if ((*count=howmny(start,prec,0)) == 0) {
          msg=margn[mnum];
          if (*(msg-1) == ':') {
               *(msg-1)='\0';
               *count=howmny(start,prec,1);
               if (mnum != margc-1) {
                    msg++;
               }
          }
          else {
               msg="";
          }
          rstrin();
     }
     else if (*count == 1) {
          rstrin();
          for (i=mnum+1 ; i < margc ; i++) {
               msg=margn[i];
               if (*(msg-1) == ':') {
                    *(msg-1)='\0';
                    if (!sameto(start,othuap->userid)) {
                         *(msg-1)=':';
                    }
                    else {
                         i++;
                    }
                    break;
               }
               else {
                    *msg='\0';
                    gobl=sameto(start,othuap->userid);
                    if (i < margc-1) {
                         *msg=' ';
                    }
               }
               if (!gobl) {
                    break;
               }
          }
          msg=(i == margc ? "" : margv[i]);
     }
     else {
          rstrin();
          if ((msg=strchr(start,':')) == NULL) {
               margn[mnum][0]='\0';
               *count=NEEDCO;
               msg="";
          }
          else {
               *msg='\0';
               if (*++msg != '\0') {
                    msg++;
               }
               *count=howmny(start,prec,1);
          }
     }
     return(msg);
}

STATIC int
howmny(stg,prec,samas)             /* number of people starting w/ stg?    */
char *stg;                              /* stg to use in sameto() test     */
int prec;                               /* how precise should this be? 0-2 */
int samas;                              /* take sameas() as a return(1)?   */
{
     int cnt=0,dstusn;

     for (othusn=0,othusp=user ; othusn < nterms ; othusn++,othusp++) {
          othuap=uacoff(othusn);
          switch (prec) {
          case 2:
               if (tlclst[othusn].channel != tlclst[usrnum].channel) {
                    break;
               }
          case 1:
               if (othusp->state != tlcstt || othusp->substt != TLKING) {
                    break;
               }
          case 0:
               if (othusp->flags&INVISB || othusp->class < SUPLON
                 || !sameto(stg,othuap->userid)) {
                    break;
               }
               if (samas && sameas(stg,othuap->userid)) {
                    return(1);
               }
               dstusn=othusn;
               cnt++;
          }
     }
     if (cnt == 1) {
          othusn=dstusn;
          othusp=&user[othusn];
          othuap=uacoff(othusn);
     }
     return(cnt);
}

void
outtlc(void)                       /* prf()s prfbuf to teleconference chan */
{
     int ousn;
     char sav,*ptr;
     struct user *ousp;

     ptr=prfptr;
     prfmlt(ETLPMT);
     sav=*ptr;
     for (ousn=0,ousp=user ; ousn < nterms ; ousn++,ousp++) {
          if (ousn != usrnum && ousp->state == tlcstt) {
               if (tlcptr->channel == tlclst[ousn].channel) {
                    if (!(tlclst[ousn].FORFLS(usrnum)&FNDBIT(usrnum))) {
                         switch (ousp->substt) {
                         case TLKING:
                              outmlt(ousn);
                              break;
                         case LINKED:
                              *ptr='\0';
                              undupe(prfbuf+findstar(prfbuf),1);
                              *ptr=sav;
                              *(ptr-1)='\0';
                              btuxmt(ousn,prfbuf+findstar(prfbuf));
                              btuxct(ousn,1,"\r");
                              *(ptr-1)='\r';
                              break;
                         }
                    }
               }
          }
     }
     clrmlt();
}

STATIC void
tlctck(void)                       /* real-time teleconference kicker      */
{
     struct tlc *tlcptr;

     for (usrnum=0,tlcptr=tlclst ; usrnum < nterms ; usrnum++,tlcptr++) {
          usrptr=&user[usrnum];
          if (usrptr->substt == FLHING) {
               usrptr->flags|=ACTIVE;
          }
          if (tlcptr->paged != 0) {
               tlcptr->paged--;
          }
          if (tlcptr->blinkc != 0) {
               if ((tlcptr->blinkc-=1) == 0) {
                    shochl(uacoff(usrnum)->userid,'',
                           baudat(user[usrnum].baud,0));
               }
          }
          if (tlcptr->reqcha != 0) {
               tlcptr->reqcha--;
          }
          tlcptr->swchan=0;
          tlcptr->flags&=~JUSTEX;
     }
     rtkick(15,tlctck);
}

STATIC void
initls(void)                       /* initialize teleconferencer list      */
{
     othusn=-1;
     othusp=user-1;
     tptr=tlclst-1;
}

STATIC void
outfgm(void)                       /* output stuff for a tele-game         */
{
     if (usrptr->substt == TLKING && *prfbuf != '\0') {
          prfmlt(ETLPMT);
          outmlt(usrnum);
          clrmlt();
     }
}

STATIC char *
tlsrui(void)                       /* teleconferencer list, show next one  */
{
     static char retval[UIDSIZ+6];

     while (othusn < nterms-1) {
          othusn++;
          othusp++;
          othuap=uacoff(othusn);
          tptr++;
          if (othusp->state == tlcstt && othusn != usrnum
             && (othusp->substt == TLKING || othusp->substt == LINKED)
             && tptr->channel == tlcptr->channel) {
               if (!(othusp->flags&INVISB)) {
                    strcpy(retval,othuap->userid);
                    if (tlcptr->FORFLS(othusn)&FNDBIT(othusn)) {
                         strcat(retval," (f)");
                    }
                    return(retval);
               }
          }
     }
     return(NULL);
}

STATIC void
tcsthn(void)                       /* teleconference status handler        */
{
     setmbk(tlcmb);
     tlcptr=&tlclst[usrnum];
     if (!isalnum(usaptr->userid[0])) {
          switch (status) {
          case CMDOK:
          case CMN2OK:
          case CM25OK:
          case 'M':
          case OUTMT:
          case 251:
          case 252:
          case 253:
               break;
          case 'D':
          case 'R':
          case 'V':
               btucmd(usrnum,"M");
               break;
          case 'B':
          case 'T':
               hupchi(LNKHUP);
               usrptr->substt=GOING;
          default:
               tlchup();
               rstchn();
          }
     }
     else if (status == STPFLA) {
          if (usrptr->substt == FLHING) {
               ret2tl(usrnum);
               tpfmsg(USRETD,pbptr->gamnam);
               tlcctx();
               prfmsg(ETLPMT);
               outprf(usrnum);
          }
     }
     else {
          dfsthn();
     }
}

STATIC void
ret2tl(who)                        /* return to talking sub-stt from flash */
int who;
{
     int savusn;

     savusn=usrnum;
     actas(who);
     lvpool();
     btucli(usrnum);
     if (!(usrptr->flags&INVISB)) {
          tpfmlt(USBACK,usaptr->userid,pbptr->gamnam);
          outtlc();
     }
     usrptr->substt=TLKING;
     actas(savusn);
}

STATIC void
tlchup(void)                       /* teleconference hang-up routine       */
{
     int i,j,othusn,chn;
     struct tlc *otlcpt;

     tlcptr=&tlclst[usrnum];
     tl2ptr=&tl2lst[usrnum];
     setmbk(tlcmb);
     telext();
     if (usrptr->state == tlcstt) {
          switch (usrptr->substt) {
          case TLKING:
          case LINKED:
               if (!(usrptr->flags&INVISB)) {
                    tpfmlt(TLCHUP,usaptr->userid);
                    outtlc();
               }
               break;
          case LNKING:
               if (usaptr->userid[0] != '(') {
                    user[tlcptr->chatch].substt=GOING;
                    btuinj(tlcptr->chatch,RING);
                    break;
               }
          case CHTING:
               hupchi(ECHHP2);
               break;
          case GOING:
               break;
          case FLHING:
               lvpool();
               break;
          case 90:
          case 91:
          case 92:
               setmem(telmisc,sizeof(struct telemisc),0);
          }
     }
     if (tlcptr->flags&EDITED) {
          setmem(&tlcdat,sizeof(struct tlcdat),0);
          strcpy(tlcdat.userid,usaptr->userid);
          tlcdat.flags=tlcptr->pflags;
          tlcdat.defchn=tlcptr->defchn;
          tlcdat.action=tlcptr->action;
          tlcdat.pagsts=tlcptr->pagsts;
          tlcdat.pagint=tlcptr->pagint;
          tlcdat.chaint=tlcptr->chaint;
          strcpy(tlcdat.theme,tlcptr->theme);
          strcpy(tlcdat.entmsg,tl2ptr->entmsg);
          strcpy(tlcdat.extmsg,tl2ptr->extmsg);
          setbtv(tbb);
          if (acqbtv(NULL,usaptr->userid,0)) {
               updbtv(&tlcdat);
          }
     }
     if (isuidc(usaptr->userid[0])) {
          chn=tlcptr->prvchn;
          for (otlcpt=tlclst,j=0 ; j < nterms ; otlcpt++,j++) {
               if (otlcpt->channel == chn && usrnum != j) {
                    otlcpt->channel=0;
                    if (user[j].state == tlcstt && user[j].substt == FLHING) {
                         ret2tl(j);
                    }
                    else if (user[j].state == tlcstt
                       && user[j].substt == TLKING) {
                         tlcptr=otlcpt;
                         i=usrnum;

                         curusr(j);
                         tpfmsg(HKICKO,usaptr->userid);
                         tlcctx();
                         prfmsg(ETLPMT);
                         outprf(usrnum);

                         if (!(usrptr->flags&INVISB)) {
                              tpfmlt(CAMEIN,usaptr->userid);
                              outtlc();
                         }

                         otlcpt=tlcptr;
                         curusr(i);
                    }
               }
               otlcpt->INVFLS(usrnum)&=~FNDBIT(usrnum);
               otlcpt->FORFLS(usrnum)&=~FNDBIT(usrnum);
          }
     }
     for (otlcpt=tlclst,othusn=0 ; othusn < nterms ; otlcpt++,othusn++) {
          if (otlcpt->chatch == usrnum) {
               otlcpt->flags&=~CCHVLD;
          }
     }
     tlcptr=&(tlclst[usrnum]);
     setmem(tlcptr,sizeof(struct tlc),0);
}

STATIC void
hupchi(xitmsg)                /* utility for terminating btuchi link */
int xitmsg;
{
     actas(tlcptr->chatch);
     lvchi(xitmsg,ENTTLC);
     prfmsg(ETLPMT);
     outprf(usrnum);
     actas(tlcptr->chatch);
}

STATIC void
actas(usrnm)                  /* act as specified user until further notice*/
int usrnm;
{
     usrnum=usrnm;
     usrptr=&user[usrnum];
     extptr=extoff(usrnum);
     clingo=extptr->lingo;
     usaptr=uacoff(usrnum);
     tlcptr=&tlclst[usrnum];
     tl2ptr=&tl2lst[usrnum];
     pbptr=fapyrs[usrnum].poolpt;
}

STATIC void
deltlc(userid)                /* delete user routine                       */
char *userid;
{
     setbtv(tbb);
     if (acqbtv(NULL,userid,0)) {
          delbtv();
     }
}

STATIC void
clstlc(void)                  /* close teleconference files for shutdown   */
{
     clsmsg(tlcmb);
     clsbtv(tbb);
}

STATIC void
echo(void)                    /* handle the "echo" command                 */
{
     if (sameas(margv[1],"on")) {
          btuech(usrnum,2);
          tpfmsg(ECHMON);
     }
     else if (sameas(margv[1],"off")) {
          btuech(usrnum,0);
          tpfmsg(ECHMOF);
     }
     else if (sameas(margv[1],"px")) {
          btuech(usrnum,1);
          tpfmsg(ECHMPX);
     }
     else {
          tpfmsg(ECHHLP);
     }
}

STATIC void
entrmsg(void)                      /* display default msg or enter msg     */
{
     if (!strlen(tl2ptr->entmsg)) {
          tpfmlt(ENTTLC,usaptr->userid);
     }
     else {
          tpfmlt(EXISKL,tl2ptr->entmsg);
     }
}

void
telext(void)                       /* telegames general exit functions     */
{
     tngexit();
     pokexit();
     bjexit();
}

STATIC int
urinv(void)                        /* can't execute command while invisible*/
{
     if (usrptr->flags&INVISB) {
          tpfmsg(URINVS);
          return(1);
     }
     return(0);
}

/*********************** ACTION COMMANDS/HANDLERS *************************/

#define ACTSIZ 12

static
struct action {
     char action[ACTSIZ];
     char type;
     int msgnum;
} *actarr;

static
int nactns,
    cactno;

#define ACTLIM 1000
#define NAPARM 6

#define TYPOAC(u)  (u+1)
#define ACTRES(u)  (u+2)
#define ACTOTH(u)  (u+3)
#define ACTAUX(u)  (u+4)
#define ACTUID(u)  (u+5)

static
char mtemp[ACTSIZ];

static
struct action *actptr;

STATIC void
action(void)                       /* handle the "action" command          */
{
     if (!hasmkey(ACTKEY)) {
          tpfmsg(ACLIVO);
          howbuy();
     }
     else if (sameas(margv[1],"on")) {
          tlclst[usrnum].flags|=ACTION;
          tpfmsg(ACTMON);
     }
     else if (sameas(margv[1],"off")) {
          tlclst[usrnum].flags&=~ACTION;
          tpfmsg(ACTMOF);
     }
     else if (sameas(margv[1],"list")) {
          tpfmsg(ACTLST);
     }
     else {
          tpfmsg(ACTSTF);
     }
}

STATIC int
chkact(void)                       /* master action handler                */
{
     char *mv0ptr;

     if (urinv()) {
          return(1);
     }
     tlcptr=&tlclst[usrnum];
     if ((tlcptr->flags&ACTION) && (isalpha(*margv[0]))
       && ((strlen(margv[0])) < ACTSIZ-1)) {
          strcpy(mtemp,margv[0]);
          for (mv0ptr=mtemp ; *mv0ptr != '\0' ; mv0ptr++) {
               *mv0ptr=tolower(*mv0ptr);
          }
          if (strlen(gacstg) && sameas(margv[0],gacstg)) {
               if (tlcptr->pflags&CANGA) {
                    generic();
                    return(1);
               }
               return(0);
          }
          if ((actptr=actbin(mtemp)) != NULL && actacc()) {
               cactno=actptr->msgnum;
               switch (actptr->type) {
               case 'S':
                    simples();
                    break;
               case 'D':
                    dobject();
                    break;
               case 'B':
                    bothsnd();
                    break;
               case 'Y':
                    yelling();
                    break;
               }
               return(1);
          }
     }
     return(0);
}

STATIC void
simples(void)                      /* simple action command handler        */
{
     tpfmlt(SACBEG);
     if (actptr->type == 'S') {
          tpfmlt(ACTOTH(cactno),usaptr->userid,thisher(usrnum));
     }
     else {
          tpfmlt(ACTAUX(cactno),usaptr->userid,thisher(usrnum));
     }
     pmlt("\r");
     outtlc();
     tpfmsg(ACTRES(cactno));
     prf("\r");
}

STATIC void
dobject(void)                      /* direct object action command handler */
{
     int secret=0,count;

     if (margc == 1) {
          tpfmsg(ACTWHO);
          return;
     }
     rstrin();
     if (sameas(margv[margc-1],"secretly")) {
          secret=1;
          margn[margc-2][0]='\0';
     }
     count=howmny(margv[1],2,1);
     if (sameas(margv[1],"all")) {
          if (secret) {
               tpfmsg(NAASEC,mtemp);
          }
          else {
               tpfmlt(SACBEG);
               tpfmlt(ACTOTH(cactno),usaptr->userid,"everyone",thisher(usrnum));
               pmlt("\r");
               outtlc();
               tpfmsg(ACTRES(cactno));
               prf("\r");
          }
     }
     else {
          if (count == 0) {
               tpfmsg(WHSNHR,margv[1]);
          }
          else if (count != 1) {
               tpfmsg(AMBIG,margv[1],mtemp);
          }
          else if (tlclst[othusn].FORFLS(usrnum)&FNDBIT(usrnum)) {
               tpfmsg(IGNORN,othuap->userid);
          }
          else {
               if (secret) {
                    tpfmlt(SACBEG);
                    tpfmlt(ACTOTH(cactno),usaptr->userid,"you",thisher(usrnum));
                    pmlt("\b... secretly!\r");
                    prfmlt(ETLPMT);
                    outmlt(othusn);
                    tpfmsg(ACTRES(cactno));
                    prf("\b... and just to %s!\r",uacoff(othusn)->userid);
               }
               else {
                    tpfmlt(SACBEG);
                    tpfmlt(ACTOTH(cactno),usaptr->userid,"you",thisher(usrnum));
                    pmlt("\r");
                    prfmlt(ETLPMT);
                    outmlt(othusn);
                    tpfmlt(SACBEG);
                    tpfmlt(ACTOTH(cactno),usaptr->userid,uacoff(othusn)->userid,
                                          thisher(usrnum));
                    pmlt("\r");
                    outbt2(othusn);
                    tpfmsg(ACTRES(cactno));
                    prf("\r");
               }
          }
     }
}

STATIC void
bothsnd(void)                      /* simple-or-direct object action handler */
{
     if (margc == 1) {
          simples();
     }
     else {
          dobject();
     }
}

STATIC void
yelling(void)                      /* "yell" type action command handler   */
{
     if (margc == 1) {
          tpfmlt(SACBEG);
          tpfmlt(ACTOTH(cactno),usaptr->userid,thisher(usrnum));
          pmlt("\r");
          outtlc();
          tpfmsg(ACTRES(cactno));
          prf("\r");
     }
     else {
          tpfmlt(ACTFRM,usaptr->userid);
          tpfmlt(ACTAUX(cactno));
          rstrin();
          tpfmlt(ACTFR2,margv[1]);
          outtlc();
          tpfmsg(AMSENT);
     }
}

STATIC void
generic(void)                      /* generic type action command handler  */
{
     if (margc == 1) {
          tpfmsg(GNSENT);
     }
     else {
          rstrin();
          tpfmlt(GACMSG,usaptr->userid,margv[1]);
          outtlc();
          tpfmsg(GASENT);
     }
}

STATIC int
actacc(void)                       /* check whether action allowed this guy */
{
     char *uidstg;

     if (*(uidstg=rawmsg(ACTUID(actptr->msgnum))) != '\0') {
          if (!sameas(usaptr->userid,uidstg) && !(tlcptr->pflags&TSYSOP)) {
               return(0);
          }
     }
     return(1);
}

STATIC void
inidya(void)                       /* init dynamic action array from msgs  */
{
     char *actstg;
     int i,j,len,n;
     char typ;

     setmbk(tlcmb);
     nactns=numopt(NACTNS,1,ACTLIM);
     actarr=(struct action *)alcmem(n=(nactns+1)*sizeof(struct action));
     setmem(actarr,n,0);
     actarr++;
     for (i=(ACTIONS+1), j=0 ; j < nactns ; j++) {
          if (*(actstg=rawmsg(i)) == '\0') {
               catastro("TLC: Action #%d null!",i-ACTIONS);
          }
          if ((len=strlen(actstg)) >= ACTSIZ) {
               catastro("TLC: Action #%d too long!",i-ACTIONS);
          }
          movmem(actstg,&actarr[j].action,len);
          typ=chropt(TYPOAC(i));
          if (typ != 'S' && typ != 'D' && typ != 'B' && typ != 'Y') {
               catastro("TLC: Action #%d -- unknown type!",i-ACTIONS);
          }
          actarr[j].type=typ;
          actarr[j].msgnum=i;
          i+=NAPARM;
     }
}

STATIC struct action *
actbin(stgptr)                     /* binary search for action word        */
char *stgptr;
{
     int cond;
     struct action *low,*mid,*high;

     low=&actarr[0];
     high=&actarr[nactns-1];
     while (low <= high) {
          mid=low+((int)(high-low))/2;
          if ((cond=strcmp(stgptr,mid->action)) < 0) {
               if (mid == low) {
                    break;
               }
               high=mid-1;
          }
          else if (cond > 0) {
               if (mid == high) {
                    break;
               }
               low=mid+1;
          }
          else {
               return(mid);
          }
     }
     return(NULL);
}

/*****************  END OF ACTION COMMANDS/HANDLERS  ***********************/

/*****************  TELECONFERENCE EDITOR SUBSTATE ROUTINES ***************/

static
char *defchs[3]={
     "","Main Channel","Private Channel"
};

static
char *onoffs[4]={
     "","ON","OFF","OK"
};

STATIC void
teditor(void)                      /* teleconference editor main menu      */
{
     if (margc == 0) {
          tedtmnu();
     }
     else {
          switch (toupper(*margv[0])) {
          case 'X':
               tpfmsg(EXIEDT);
               outprf(usrnum);
               usrptr->substt=TLKING;
               btumil(usrnum,tinpsz);
               if (!(usrptr->flags&INVISB)) {
                    entrmsg();
                    outtlc();
               }
               tpfmsg(INTRO);
               tlcctx();
               tlcptr->flags|=EDITED;
               break;
          case '1':
               tpfmsg(DEFCHN);
               usrptr->substt=11;
               break;
          case '2':
               tpfmsg(DEFACT);
               usrptr->substt=12;
               break;
          case '3':
               tpfmsg(PAGOOO);
               usrptr->substt=13;
               break;
          case '4':
               tpfmsg(PAGINT);
               usrptr->substt=14;
               break;
          case '5':
               tpfmsg(CHAINT);
               usrptr->substt=15;
               break;
          case '6':
               tpfmsg(CHATHE);
               btumil(usrnum,40);
               usrptr->substt=16;
               break;
          default:
               tpfmsg(SAYWHA);
               tedtmnu();
          }
     }
}

STATIC void
tedtmnu(void)                      /* display current user settings        */
{
     usrptr->substt=EDTING;
     prfmsg(EDTMNU,(!strlen(tl2ptr->entmsg) ? "(none)" : tl2ptr->entmsg),
                   (!strlen(tl2ptr->extmsg) ? "(none)" : tl2ptr->extmsg),
                   defchs[tlcptr->defchn],
                   onoffs[tlcptr->action],
                   onoffs[tlcptr->pagsts],
                   tlcptr->pagint/4,
                   tlcptr->chaint/4,
                   (tlcptr->theme[0] == '\0' ? "(none)" : tlcptr->theme));
}

STATIC void
edtr01(void)                       /* handle editing default channel       */
{
     if (margc == 0) {
          tedtmnu();
     }
     else {
          switch (*margv[0]) {
          case '1':
               tpfmsg(OKSET);
               tlcptr->defchn=1;
               tedtmnu();
               break;
          case '2':
               tpfmsg(OKSET);
               tlcptr->defchn=2;
               tedtmnu();
               break;
          default:
               tpfmsg(SAYWHA);
               tpfmsg(DEFCHN);
          }
     }
}

STATIC void
edtr02(void)                       /* handle setting action mode on/off    */
{
     if (margc == 0) {
          tedtmnu();
     }
     else {
          switch (*margv[0]) {
          case '1':
               tpfmsg(OKSET);
               tlcptr->action=1;
               tedtmnu();
               break;
          case '2':
               tpfmsg(OKSET);
               tlcptr->action=2;
               tedtmnu();
               break;
          default:
               tpfmsg(SAYWHA);
               tpfmsg(DEFACT);
          }
     }
}

STATIC void
edtr03(void)                       /* handle setting page status           */
{
     if (margc == 0) {
          tedtmnu();
     }
     else {
          switch (*margv[0]) {
          case '1':
               tpfmsg(OKSET);
               tlcptr->pagsts=1;
               tedtmnu();
               break;
          case '2':
               tpfmsg(OKSET);
               tlcptr->pagsts=2;
               tedtmnu();
               break;
          case '3':
               tpfmsg(OKSET);
               tlcptr->pagsts=3;
               tedtmnu();
               break;
          default:
               tpfmsg(SAYWHA);
               tpfmsg(PAGOOO);
          }
     }
}

STATIC void
edtr04(void)                       /* handle page interval                 */
{
     int i;

     if (margc == 0) {
          tedtmnu();
     }
     else if ((i=atoi(margv[0])) > 0 && i < 10) {
          tpfmsg(OKSET);
          tlcptr->pagint=(4*i);
          tedtmnu();
     }
     else {
          tpfmsg(SAYWHA);
          tpfmsg(PAGINT);
     }
}

STATIC void
edtr05(void)                       /* handle chat intervals                */
{
     int i;

     if (margc == 0) {
          tedtmnu();
     }
     else if ((i=atoi(margv[0])) > 0 && i < 10) {
          tpfmsg(OKSET);
          tlcptr->chaint=(4*i);
          tedtmnu();
     }
     else {
          tpfmsg(SAYWHA);
          tpfmsg(CHAINT);
     }
}

STATIC void
edtr06(void)                       /* handle setting a channel topic       */
{
     if (margc == 0) {
          tlcptr->theme[0]='\0';
          tedtmnu();
     }
     else {
          rstrin();
          strcpy(tlcptr->theme,margv[0]);
          tpfmsg(OKSET);
          tedtmnu();
     }
}

/*************** END OF TELECONFERENCE EDITING SUBSTATE ROUTINE ************/

/*****************  SYSOP EDITOR FOR ENTER/EXIT MESSAGES    ***************/

static
struct tlcdat othtlc;

STATIC void
usredt(void)                       /* entering sysop editor state          */
{
     setmem(telmisc,sizeof(struct telemisc),0);
     usrptr->substt=90;
     tpfmsg(SYSEDT);
     btumil(usrnum,UIDSIZ-1);
     telmisc->edtin=1;
}

STATIC void
chkhdl(void)                       /* select a user to edit or exit editor */
{
     if (margc == 0) {
          usredt();
     }
     else if (sameas(margv[0],"x")) {
          setmem(telmisc,sizeof(struct telemisc),0);
          tpfmsg(EXIEDT);
          outprf(usrnum);
          usrptr->substt=TLKING;
          btumil(usrnum,tinpsz);
          if (!(usrptr->flags&INVISB)) {
               entrmsg();
               outtlc();
          }
          tpfmsg(INTRO);
          tlcctx();
     }
     else {
          rstrin();
          makhdl(input);
          setbtv(tbb);
          if (qeqbtv(input,0)) {
               gcrbtv(&othtlc,0);
               movmem(input,telmisc->uname,UIDSIZ);
               pmtedw();
               usrptr->substt=91;
          }
          else {
               tpfmsg(NOSUCH);
          }
     }
}

STATIC void
pmtedw(void)                       /* prompt for "edit which field?"       */
{
     struct usracc tmpacc;

     if (!onsysn(telmisc->uname,1)) {
          setbtv(accbb);
          geqbtv(othuap=&tmpacc,telmisc->uname,0);
     }
     prfmsg(EDTWC2,telmisc->uname,
          (!strlen(othtlc.entmsg) ? "<None Set>" : othtlc.entmsg),
          (!strlen(othtlc.extmsg) ? "<None Set>" : othtlc.extmsg),
          ((othtlc.flags&TSYSOP) ? "Yes" : "No"),
          ((othtlc.flags&CANGA) ? "Yes" : "No"),
          l2as(othuap->creds));
}

STATIC void
edtwch(void)                       /* select choice to edit or exit out    */
{
     if (margc == 0) {
          pmtedw();
     }
     else {
          switch (toupper(*margv[0])) {
          case 'X':
               setbtv(tbb);
               if (qeqbtv(telmisc->uname,0)) {
                    dedcds();
               }
               else {
                    prfmsg(RECNUP);
               }
               usredt();
               break;
          case '1':
          case '2':
               telmisc->wchmsg=atoi(margv[0]);
               shomsg();
               break;
          case '3':
               updflag(TSYSOP);
               break;
          case '4':
               updflag(CANGA);
               break;
          default:
               tpfmsg(SAYWHA);
               pmtedw();
          }
     }
}

STATIC void
updflag(flag)                      /* update one of the users flags        */
int flag;
{
     if (othtlc.flags&flag) {
          othtlc.flags&=~flag;
     }
     else {
          othtlc.flags|=flag;
     }
     telmisc->updrec|=UPDFLG;
     pmtedw();
}

STATIC void
shomsg(void)                       /* show current message, prompt for new */
{
     prfmsg(SHOSTF,telmisc->wchmsg == 1 ? "entrance" : "exit",
          telmisc->wchmsg == 1 ? (!strlen(othtlc.entmsg) ? "<None Set>" : othtlc.entmsg) :
                                 (!strlen(othtlc.extmsg) ? "<None Set>" : othtlc.extmsg));
     usrptr->substt=92;
     btumil(usrnum,TMSSIZ-1);
}

STATIC void
updmsg(void)                       /* update the message in the work area  */
{
     char *msgptr;

     if (margc) {
          msgptr=(telmisc->wchmsg == 1 ? othtlc.entmsg : othtlc.extmsg);
          setmem(msgptr,TMSSIZ,0);
          if (margc == 1 && toupper(*margv[0]) == 'X') {
          }
          else {
               rstrin();
               strcpy(msgptr,input);
               if (telmisc->wchmsg == 1) {
                    telmisc->entchg=1;
               }
               else {
                    telmisc->extchg=1;
               }
          }
          telmisc->updrec|=UPDMSG;
     }
     usrptr->substt=91;
     btumil(usrnum,tinpsz);
     pmtedw();
}

STATIC void
dedcds(void)                       /* deduct the credits from the user     */
{
     int chgamt;
     struct usracc tmpacc;

     if (telmisc->updrec&UPDFLG) {
          if (onsysn(telmisc->uname,1)) {
               tlclst[othusn].pflags=0;
               tlclst[othusn].pflags|=othtlc.flags&(TSYSOP+CANGA);
               tlclst[othusn].flags|=EDITED;
          }
     }
     if (telmisc->updrec&UPDMSG) {
          if (oktocg) {
               chgamt=(telmisc->entchg+telmisc->extchg)*msgchg;
               if (chgamt) {
                    if (onsysn(telmisc->uname,1)) {
                         if (!odedcrd(othusn,(long)chgamt,0,0)) {
                              tpfmsg(CNTAFT,telmisc->uname);
                              return;
                         }
                         else {
                              movmem(othtlc.entmsg,tl2lst[othusn].entmsg,TMSSIZ);
                              movmem(othtlc.extmsg,tl2lst[othusn].extmsg,TMSSIZ);
                              tlclst[othusn].flags|=EDITED;
                              clrprf();
                              tpfmlt(YOUUPD2);
                              injoth();
                              tpfmsg(USRCHD2,telmisc->uname);
                         }
                    }
                    else {
                         setbtv(accbb);
                         if (acqbtv(&tmpacc,telmisc->uname,0)) {
                              if (!ndedcrd(tmpacc.userid,(long)chgamt,0,0)) {
                                   tpfmsg(CNTAFT,telmisc->uname);
                                   return;
                              }
                              else {
                                   tpfmsg(USRDED,telmisc->uname,chgamt);
                              }
                         }
                         else {
                              catastro("TLC: No master for %s!",telmisc->uname);
                         }
                    }
                    if (dispad) {
                         shocst("TELECONFERENCE CREDIT CHARGE",
                                "%s charged %d credits",telmisc->uname,chgamt);
                    }
               }
               else {
                    chgmsg();
               }
          }
          else {
               chgmsg();
          }
     }
     setbtv(tbb);
     gcrbtv(NULL,0);
     updbtv(&othtlc);
     prfmsg(URCUPD);
}

STATIC void
chgmsg(void)                       /* enter/exit changed no charge update  */
{
     if (onsys(telmisc->uname)) {
          movmem(othtlc.entmsg,tl2lst[othusn].entmsg,TMSSIZ);
          movmem(othtlc.extmsg,tl2lst[othusn].extmsg,TMSSIZ);
          tlclst[othusn].flags|=EDITED;
          clrprf();
          tpfmlt(YOUUPDF);
          injoth();
          tpfmsg(USRCHDF,telmisc->uname);
     }
}
/*************** END OF SYSOP ENTER/EXIT MESSAGE ***************************/

