/****************************************************************************
*                                                                           *
*   GALAED.C                                                                *
*                                                                           *
*   Copyright (C) 1992-1993 GALACTICOMM, Inc.   All rights reserved.        *
*                                                                           *
*   This utility allows the editing, deletion and creation of ACTION        *
*   TELECONFERENCE actions.  It will create a new GALETL.MSG containing the *
*   list of actions in the proper format.                                   *
*                                                                           *
*                                        - Bill Hyatt  11/92                *
*                                                                           *
****************************************************************************/

#include "gcomm.h"
#include "usracc.h"
#include "galsef.h"

/*----- various actions-related -------------------------------------------*/
#define MAXACT 600                   /* maximum # of actions allowed       */
#define ACTNAM 0                     /* action name - edit field #0        */
#define ACTTYP 1                     /*        type              #1        */
#define ACTRES 2                     /*        resp. to user     #2        */
#define ACTOTH 3                     /*        othr usrs see     #3        */
#define ACTAUX 4                     /*        aux message       #4        */
#define ACTUID 5                     /*        spec usr id       #5        */
#define NAMLEN 10                    /* action name max len                */
#define TYPLEN 6                     /* action type len                    */
#define RSPLEN 78                    /* len of various teleconf resp's     */
#define DSPLEN 65                    /* max field display width            */
#define AINWIN 23                    /* # action names in choose window    */
#define SIMPLE 0                     /* /--------- action types ---------\ */
#define DIRECT 1                     /*                 see                */
#define BOTH   2                     /*        ACTION TELECONFERENCE       */
#define YELL   3                     /* \------------- docs -------------/ */
#define NOCL   0                     /* bgclsq(): actn has no ANSI clr str */
#define SHRTCL 10                    /*           short ANSI color string  */
#define MIDLCL 12                    /*           mid-length  "       "    */
#define LONGCL 14                    /*           long        "       "    */
#define NEW    0                     /* save new action (savact, etc.)     */
#define EXISTS 1                     /* save existing action (savact, etc) */
#define DELETE 2                     /* delete action (savact, etc.)       */
#define ACTPLN 6                     /* # action words/line in ACTLST      */
#define ACTWLM 126                   /* max # actions written to ACLTST    */
/*----- results from edtfld() & chstyp() ----------------------------------*/
#define ABNACT 0                     /* user wants to abandon an action    */
#define SAVACT 1                     /* user wants to save an action       */
#define NXTFLD 2                     /* go to next field to edit           */
#define PRVFLD 3                     /* go to prev field to edit           */
#define RSTFLD 4                     /* restore, then re-edit field        */
#define CHSCOL 5                     /* choose color, then re-edit field   */
/*----- results from chkfld() ---------------------------------------------*/
#define FLDOK  0                     /* no errors in field                 */
#define TMNYPS 1                     /* too many "%s" constructs           */
#define TFEWPS 2                     /* too few "%s" constructs            */
#define ILGCST 3                     /* illegal construct (eg. "%f")       */
#define NONAME 4                     /* no name given to action            */
#define NOLEN  5                     /* nothing entered for req. field     */
#define NOBRAK 6                     /* YELL type, no "<>" in "Aux. Msg."  */
#define NOLEFB 7                     /* YELL type, no "<"                  */
#define NORITB 8                     /* YELL type, no ">"                  */
#define NMEXST -99                   /* action name already exists         */
/*----- screens-related ---------------------------------------------------*/
#define NSCNS  5                     /* # of screens in pre-OBJ'd table    */
#define NFLDS  6                     /* # of fields on screen              */
#define BASE   0                     /* overall base screen                */
#define PTTYPH 1                     /* action point/action type help      */
#define FLDHLP 2                     /* field help                         */
#define MSCOBJ 3                     /* misc. objects/aux help msgs        */
#define COLORB 4                     /* choose colors box                  */
#define HLPUPX 0                     /* help msg "from" upx                */
#define HLNLEN 65                    /* len of line in help msg            */
#define NHLPLN 5                     /* # lines in help msg                */
/*----- screen colors -----------------------------------------------------*/
#define DSPATR 0x1E                  /* field display, no ansi color       */
#define BLKATR 0x11                  /* blank field before display         */
#define MSSATR 0x4F                  /* messages to user                   */
#define YNCATR 0x0F                  /* yes/no ques. / choose color        */
#define NCEATR 0x0E                  /* field edit, no ansi color          */
#define NAMATR 0x1F                  /* display field name                 */
#define CRNATR 0x1B                  /* highlight current fld name         */
#define DIMATR 0x17                  /* temp dim something on screen       */
/*----- miscellaneous -----------------------------------------------------*/
#define EDTFLD 0                     /* key help when editing fields       */
#define EDTTYP 6                     /* key help when choosing type        */
#define RSFHLP 3                     /* vary key help so ESC=restore fld   */
#define WTHSVE 1                     /* DOS exit code for save and exit    */
#define WOTSVE 0                     /* DOS exit code for exit w/out save  */
#define REDISP -1                    /* need to re-display choice          */
#define SAME   -1                    /* re-edit current action             */
#define NEWACT -9999                 /* new action                         */
#define CURACT -9998                 /* current sequential action          */
#define PUNCHR " .,'-"               /* punc. chars in user id's           */
#define DEFCOL 0x0A                  /* default color for new actions      */
#define ANYPOS -1                    /* rdfile(): scan for stg anywhere    */
#define EOLST  -1                    /* chgptrs(): action goes at list end */
#define NONE   -1                    /* not present (misc. circumstances)  */
#define MXBFSZ 256                   /* assumed maximum stg buffer size    */
#define MINMEM 10000L                /* min memory before we warn op       */

extern int fstchc,                   /* 1st guy in select window           */
           curchc;                   /* currently selected guy in window   */

extern char scntbl[][SCNSIZ];        /* table of pre-OBJ'd scn files       */

struct action {                      /* action data block                  */
     int  msgnum;                    /*      beg msg# for info from .mcv   */
     char type;                      /*      action type                   */
     int  color;                     /*      ANSI colors for action        */
     char **chginf;                  /*      pointer to new/changed info   */
     struct action *nxtact;          /*      next action in list           */
} *actlst,                           /* list of actions                    */
  *aclstl,                           /* last action in list                */
  *tmp;                              /* utility/work pointer               */

struct actinf {                      /* action info data block             */
     char actnam[NAMLEN+1];          /*      action name                   */
     char acttyp;                    /*      action type                   */
     char actres[RSPLEN+1];          /*      resp to user who typed action */
     char actoth[RSPLEN+1];          /*      what other users see          */
     char actaux[RSPLEN+1];          /*      auxiliary message             */
     char actuid[UIDSIZ];            /*      specific user id              */
     int  color;                     /*      ANSI colors for action        */
} tmpinf,                            /* temp info for edit/display         */
  orginf;                            /* info before edit for comparison    */

struct field {                       /* field screen info data block       */
     char *contents;                 /*      contents of field             */
     char *ocontnts;                 /*      contents before edit          */
     char *name;                     /*      field name                    */
     int  nmlocx,nmlocy;             /*      name location on screen       */
     int  blklen;                    /*      len of blk line to print      */
     int  edtlocx,edtlocy;           /*      edit/disp loc on screen       */
     int  edtlen,edwnsz;             /*      buffer size, edit win length  */
     int  hlpfil,hlpupy;             /*      help msg: scn, "from" upy     */
     int  ahlpfil,ahlpupy;           /*      aux help msg, if necc.        */
} fields[]={                         /* array of field info                */
     {tmpinf.actnam,orginf.actnam,"Name:",
           13,2,
           10,
           19,2,
           NAMLEN+1,NONE,
           FLDHLP,0,
           NONE,NONE
     },
     {NULL,NULL,"Type:",
           46,2,
           6,
           52,2,
           NONE,
           NONE,NONE,
           NONE,NONE
     },
     {tmpinf.actres,orginf.actres,"Response To User:",
           13,4,
           65,
           13,5,
           RSPLEN+1,DSPLEN,
           FLDHLP,5,
           NONE,NONE
     },
     {tmpinf.actoth,orginf.actoth,"Other Users See:",
           13,7,
           65,
           13,8,
           RSPLEN+1,DSPLEN,
           FLDHLP,10,
           MSCOBJ,15
     },
     {tmpinf.actaux,orginf.actaux,"",
           13,10,
           65,
           13,11,
           RSPLEN+1,DSPLEN,
           FLDHLP,15,
           MSCOBJ,20
     },
     {tmpinf.actuid,orginf.actaux,"Specific User-ID:",
           13,13,
           29,
           31,13,
           0,NONE,
           FLDHLP,20,
           NONE,NONE
     }
};

int typhlp[]={5,10,15,20};           /* array of upy's for type-help msgs  */

int asct[8]={0,4,2,6,1,5,3,7};       /* ANSI atr/chr translation table     */
char ansseq[LONGCL+1];               /* ANSI char seq to write to file     */

char **anmlst;                       /* array of action names 4 choose win */

char *actyps[]={                     /* array of action types              */
     "Simple",
     "Direct",
     "Both",
     "Yell"
};

char *flderr[]={                     /* array of chkfld() error messages   */
     NULL,
     "This field cannot have %s \"%%s\" constructs!...",
     "This field must have at least %d \"%%s\" construct%s!...",
     "Only \"%s\" and/or \"%%\" allowed in action!...",
     "Action must have a name!...",
     "Something must be entered in this field!...",
};

char line[MXBFSZ+1];                 /* text buffer for file reads         */

FILE *galetl;                        /* "GALETL.MCV"                       */

struct scninfo {                     /* saved screen info data block       */
     char *scn;                      /*      saved screen                  */
     int  color;                     /*      fields' colors when saved     */
     struct scninfo *nxtscn;         /*      next screen on stack          */
} *screens,                          /* LIFO stack of saved screens        */
  *curscn;                           /* pointer to top of screen stack     */

int nacmsg,                          /* msg number for "NACTNS"            */
    cmsgnm,                          /* current message in GALETL.MCV      */
    nactns=0,                        /* current # of actions               */
    aeddon=0,                        /* done with actions editor?          */
    chgsmd=0,                        /* any unsaved changes?               */
    fldchg,                          /* changed anything in pres field?    */
    edtnga=0,                        /* editing a specific action?         */
    savit=1,                         /* save this action?                  */
    ynvalid,                         /* kyhdlr(): validating  yes/no resp? */
    chsngtyp=0,                      /* choosing an action's type?         */
    chsngcol=0,                      /* choosing an action's colors?       */
    ckpsax=1,                        /* check for "%s" in "Aux. Message"?  */
    curclr,                          /* last selected color                */
    digalw,                          /* digits allowed in user id?         */
    fulalw,                          /* punc. chars allowed in user id?    */
    fnum,                            /* field currently editing            */
    edtclr,                          /* color of field being edited        */
    typidx,                          /* currently selected action type     */
    minp_s,                          /* min # of "%s" allowed in field     */
    maxp_s;                          /* max # of "%s" allowed in field     */

void init();
void chkfiles(void);
int  msgnum(FILE *hfile,char *optnam);
void addact(void);
void delact(int actnum);
void svncon();
void extaed(int save);
void chscol(void);
void edtact(int actnum);
void dimchc(void);
int  edtfld(int fnum);
int  chstyp(void);
void chkact(void);
void chkbrc(void);
int  chkfld(int fnum);
int  chkp_s(int fnum);
void edterr(int fnun,int error);
void savact(int actnum,int status);
void chgptrs(int actnum,int pos,int status);
void savfld(char *contents,int fnum,int len);
int  memlow(void);
void auxdsc(void);
void dspact(void);
void dspfld(int fldnum);
void dsphlp(void);
void dspnac(void);
void dspnam(int fldnum,int attr);
int  fldclr(int fnum,int st4dsp);
int  fndpos(char *nam);
struct action *fndact(int actnum);
void indchg(void);
void keyhlp(int n);
int  kyhdlr(int c,char *sval);
void ldtmpa(int actnum);
void getfld(char *dest,char *src,int fnum,int len);
int  mssage(char *mlin,int ynmsg,char *yndef,int pause);
int  mychoo(int n,char *chcs[],int upx,int upy,int lox,int loy,int escok);
void pushscn(void);
void popscn(void);
void rdacts(void);
int  bgclsq(char *seq);
void rdfile(FILE *frmfle,FILE *tofle,char *endstr,int pos,int wrends);
void putline(FILE *fp);
void rszlst(int newsize);
void sttidx(void);
int  wracts(int exitng);
void atr2nms(int attr,int *fore,int *back);
void atr2str(int attr);
int  nms2atr(int fore,int back);
int  str2atr(char *seq,int seqtyp);

main(void)                           /* main program entry point           */
{
     init();
     do {
          switch (mychoo(nactns,anmlst,1,1,10,23,0)) {
          case -INS:
               if (nactns < MAXACT) {
                    if (!memlow()) {
                         addact();
                    }
               }
               else {
                    mssage(spr("No more than %d actions allowed!...",MAXACT),
                           0,"",1);
               }
               break;
          case -DEL:
               delact(curchc);
               break;
          case -ALT_S:
               svncon();
               break;
          case -F10:
               extaed(WTHSVE);
               break;
          case -ESC:
          case -F9:
               extaed(WOTSVE);
               break;
          case -ALT_C:
               chscol();
               break;
          case REDISP:
               ldtmpa(curchc);
               dspact();
               break;
          default:
               if (!memlow()) {
                    edtact(curchc);
                    if (savit) {
                         savact(curchc,EXISTS);
                    }
                    else {
                         ldtmpa(curchc);
                    }
                    dspact();
               }
          }
     } while (!aeddon);
     return(0);
}

void
init(void)                           /* intialize/set up various things    */
{
     register int i;
     FILE *hfile,*bbssup;

     chkfiles();
     selatr=0x3E;
     nslatr=0x1B;
     inimsg(2048);
     bbssup=opnmsg("BBSSUP.MCV");
     setmbk(bbssup);
     if ((hfile=fopen("BBSSUP.H",FOPRA)) == NULL) {
          catastro("CANNOT FIND \"BBSSUP.H\"");
     }
     fields[ACTUID].edtlen=numopt(msgnum(hfile,"SGNUSZ"),3,29)+1;
     digalw=ynopt(msgnum(hfile,"DIGALW"));
     fulalw=ynopt(msgnum(hfile,"FULALW"));
     clsmsg(bbssup);
     fclose(hfile);
     curclr=DEFCOL;
     monorcol();
     for (i=0 ; i < NSCNS ; i++) {
          cvtscn(scntbl[i]);
     }
     screens=NULL;
     pushscn();
     cursiz(NOCURS);
     explode(scntbl[BASE],0,0,79,24);
     proff(0,0);
     explodem=0;
     rdacts();
     ldtmpa(0);
     dspact();
     dsphlp();
}

void
chkfiles(void)                       /* make sure needed files are present */
{
     struct fndblk msgfbptr,mcvfbptr,hfbptr,msxfbptr;

     if (!fnd1st(&msgfbptr,"GALETL.MSG",0)) {
          catastro("UNABLE TO FIND \"GALETL.MSG\"!");
     }
     else if (msgfbptr.attr&FAMRON) {
          strcpy(line,"FILE \"GALETL.MSG\" IS READ-ONLY!\n");
          strcat(line,"You must clear this before running the editor.");
          catastro(line);
     }
     if (fnd1st(&msxfbptr,"BBSMSX.EXE",0)) {
          if (!fnd1st(&mcvfbptr,"GALETL.MCV",0)
              || !fnd1st(&hfbptr,"GALETL.H",0)
              || msgfbptr.date > mcvfbptr.date
              || msgfbptr.date > hfbptr.date
              || (msgfbptr.date == mcvfbptr.date
                  && msgfbptr.time > mcvfbptr.time)
              || (msgfbptr.date == hfbptr.date
                  && msgfbptr.time > hfbptr.time)) {
                    system("bbsmsx galetl");
           }
     }
     else {
          catastro("UNABLE TO FIND \"BBSMSX.EXE\"!");
     }
}

int
msgnum(                              /* scan an .H file to find a msg #    */
FILE *hfile,                         /*      name of .H file to scan       */
char *optnam)                        /*      option name to look for       */
{
     int mnum;

     if (fseek(hfile,0L,0) != 0) {
          catastro("ERROR RESETTING \".H\" FILE PTR - opt=%s",optnam);
     }
     rdfile(hfile,NULL,optnam,ANYPOS,0);
     sscanf(line,"%*s %*s %d",&mnum);
     return(mnum);
}

void
addact(void)                         /* add an action to list              */
{
     edtact(NEWACT);
     if (savit) {
          if (nactns >= AINWIN) {
               rszlst(nactns+1);
          }
          tmp=(struct action *)alczer(sizeof(struct action));
          tmp->msgnum=NONE;
          tmp->chginf=(char **)alczer(NFLDS*(sizeof(char *)));
          savact(NONE,NEW);
          nactns++;
          dspnac();
     }
     else {
          ldtmpa(curchc);
     }
     dspact();
}

void
delact(                         /* delete action from list                 */
int actnum)                     /*      action to delete                   */
{
     if (nactns > 1) {
          if (mssage("Delete this action",1,"Yes",1)) {
               free(anmlst[actnum]);
               if (actnum != nactns-1) {
                    movmem(&anmlst[actnum+1],&anmlst[actnum],
                           sizeof(char *)*(nactns-actnum-1));
                    chgptrs(actnum,NONE,DELETE);
               }
               else {
                    if (actnum == fstchc) {
                         fstchc-=AINWIN;
                         if (fstchc < 0) {
                              fstchc=0;
                         }
                    }
                    curchc--;
               }
               if (nactns > AINWIN) {
                    rszlst(nactns-1);
                    if (fstchc >= nactns-AINWIN) {
                         fstchc--;
                    }
               }
               else {
                    anmlst[nactns-1]="";
               }
               nactns--;
               dspnac();
               ldtmpa(curchc);
               dspact();
               chgsmd=1;
          }
     }
     else {
          mssage("You must have at least 1 action!...",0,"",1);
     }
}

void
svncon(void)                         /* save action list and continue      */
{
     int sfst,scur;
     register int i,j;
     struct action *ttmp;

     if (wracts(0)) {
          for (i=0,tmp=actlst,actlst=NULL,aclstl=NULL ; i < nactns ; i++) {
               free(anmlst[i]);
               if (tmp->chginf != NULL) {
                    for (j = ACTNAM ; j < NFLDS ; j++) {
                         if (j != ACTTYP) {
                              if (tmp->chginf[j] != NULL) {
                                   free(tmp->chginf[j]);
                              }
                         }
                    }
               }
               ttmp=tmp;
               tmp=tmp->nxtact;
               free(ttmp);
          }
          free(anmlst);
          sfst=fstchc;
          scur=curchc;
          rdacts();
          fstchc=sfst;
          ldtmpa(curchc=scur);
          dspact();
          chgsmd=0;
     }
}

void
extaed(                              /* actions editor exit routine        */
int save)                            /*     1:save & exit  0:w/out save    */
{
     if (save ? wracts(1)
              : mssage((chgsmd ? "Changes have been made - quit anyway"
                                : "Quit Actions Editor"),
                        1,"No ",1)) {
          rstcur();
          cursiz(LILCURS);
          popscn();
          clsmsg(galetl);
          ansion(1);
          printf("\33[0m\n");
          aeddon=1;
    }
}

void
chscol(void)                         /* set ANSI colors for an action      */
{
     int fore,back,atoval;
     char val[3];
     register int i;

     chsngcol=1;
     pushscn();
     explodeto(scntbl[COLORB],0,0,56,16,18,4);
     proff(0,0);
     locate(39,4);
     ansion(1);
     if (fnum == ACTNAM && edtnga) {
          depad(fields[ACTNAM].contents);
     }
     printf("\33[0;30;47m%s\":",fields[ACTNAM].contents);
     setatr(tmpinf.color);
     atr2nms(tmpinf.color,&fore,&back);
     locate(38,14);
     printf("  foreground=%-2d   ",fore);
     locate(38,15);
     printf("  background=%-2d   ",back);
     val[0]='\0';
     while (edtval(59,18,3,val,kyhdlr,0)) {
          if (val[0] == '\0') {
               break;
          }
          if ((atoval=atoi(val)) >= 0 && atoval < 32) {
               fore=atoval;
               break;
          }
          printf("\7");
     }
     val[0]='\0';
     while (edtval(59,19,2,val,kyhdlr,0)) {
          if (val[0] == '\0') {
               break;
          }
          if ((atoval=atoi(val)) >= 0 && atoval < 8) {
               back=atoval;
               break;
          }
          printf("\7");
     }
     popscn();
     tmpinf.color=curclr=nms2atr(fore,back);
     for (i=ACTRES ; i <= ACTAUX ; i++) {
          dspfld(i);
     }
     if (!edtnga) {
          if (curclr != tmp->color) {
               tmp->color=tmpinf.color;
               chgsmd=1;
          }
          cursiz(NOCURS);
     }
     chsngcol=0;
}

void
edtact(                              /* edit fields for an action          */
int actnum)                          /*      action to edit                */
{
     dimchc();
     if (actnum == NEWACT) {
          ldtmpa(NEWACT);
          dspact();
     }
     else {
          ldtmpa(actnum);
          movmem(&tmpinf,&orginf,sizeof(struct actinf));
     }
     edtnga=1;
     savit=1;
     fnum=ACTNAM;
     pushscn();
     keyhlp(EDTFLD);
     dsphlp();
     do {
          switch (fnum != ACTTYP ? edtfld(fnum) : chstyp()) {
          case ABNACT:
               savit=0;
          case SAVACT:
               edtnga=0;
               break;
          case NXTFLD:
               dspfld(fnum);
               fnum=(fnum < ACTUID ? fnum+1 : ACTNAM);
               if (fnum == ACTAUX && typidx < BOTH) {
                    fnum=ACTUID;
               }
               break;
          case PRVFLD:
               dspfld(fnum);
               fnum=(fnum > ACTNAM ? fnum-1 : ACTUID);
               if (fnum == ACTAUX && typidx < BOTH) {
                    fnum=ACTOTH;
               }
               break;
          case CHSCOL:
               chscol();
          case RSTFLD:
               break;
          }
     } while (edtnga);
     if (savit) {
          chkact();
     }
     popscn();
     auxdsc();
}

void
dimchc(void)                         /* temp dim choices in choose window  */
{
     int i,id;

     setatr(DIMATR);
     for (i=0,id=fstchc ; i < AINWIN ; i++,id++) {
          locate(1,1+i);
          printf("%-10s",anmlst[id]);
     }
}

int
edtfld(                              /* edit individual field              */
int fnum)                            /*      field to edit                 */
{
      int edtngf=1,strpl,rc;
      struct field *curfld;
      static int oedtclr,oedtedw,oedtxof,oedtbof,edtsam=0;

      dsphlp();
      dspnam(fnum,CRNATR);
      keyhlp(EDTFLD);
      curfld=fields+fnum;
      supedt(curfld->edtlocx,curfld->edtlocy,curfld->edtlen,curfld->contents,
             kyhdlr,MULTIEX);
      if (!edtsam) {
           if (curfld->edwnsz != NONE) {
                edtedw=curfld->edwnsz;
                if ((strpl=strpln(curfld->contents)) == edtedw) {
                     edtxof=edtedw;
                }
                else {
                     edtxof=(strpl < edtedw ? strpl : 0);
                }
           }
      }
      else {
           edtclr=oedtclr;
           edtedw=oedtedw;
           edtxof=oedtxof;
           edtbof=oedtbof;
           edtsam=0;
      }
      edtclr=fldclr(fnum,0);
      fldchg=0;
      do {
          rfredt(edtclr);
          switch (edtvalc=getchc()) {
          case F10:
               edtngf=0;
               rc=SAVACT;
               break;
          case F9:
               edtvalc=ESC;
               fldchg=0;
          case ESC:
               if (fldchg) {
                    edtbof=edtxof=0;
                    fldchg=0;
                    keyhlp(EDTFLD);
                    rc=RSTFLD;
               }
               else {
                    rc=ABNACT;
               }
               edtngf=0;
               hdledt();
               break;
          case TAB:
          case '\r':
               edtngf=0;
               rc=NXTFLD;
               break;
          case BAKTAB:
               edtngf=0;
               rc=PRVFLD;
               break;
          case CRSRUP:
          case CRSRDN:
               break;
          case '\b':
          case DEL:
               indchg();
          case CRSRLF:
          case CRSRRT:
               hdledt();
               break;
          case ALT_C:
               oedtclr=edtclr;
               oedtedw=edtedw;
               oedtxof=edtxof;
               oedtbof=edtbof;
               edtsam=1;
               edtngf=0;
               rc=CHSCOL;
               break;
          default:
               indchg();
               edtngf=hdledt();
          }
     } while (edtngf);
     endedt();
     if (fnum == ACTNAM) {
          strlwr(fields[ACTNAM].contents);
          dspfld(ACTNAM);
     }
     dspnam(fnum,NAMATR);
     return(rc);
}

int
chstyp(void)                         /* choose action type from menu       */
{
     int inloop=1,key,otidx,sfst,scur,rc;

     chsngtyp=1;
     sfst=fstchc;
     scur=curchc;
     pushscn();
     dspnam(fnum,CRNATR);
     fldchg=0;
     dsphlp();
     keyhlp(EDTTYP);
     nsexploto(scntbl[MSCOBJ],0,0,7,5,59,1);
     proff(0,0);
     fstchc=0;
     curchc=otidx=typidx;
     do {
          switch(key=mychoo(4,actyps,60,2,65,5,1)) {
          case -ESC:
               if (fldchg) {
                    sttidx();
                    curchc=typidx;
               }
               else {
                    inloop=0;
                    break;
               }
          case -INS:
          case -DEL:
          case -ALT_S:
          case REDISP:
               typidx=curchc;
               if (key != -ESC && typidx != otidx) {
                    fldchg=1;
                    keyhlp(EDTTYP+RSFHLP);
               }
               else {
                    fldchg=0;
                    keyhlp(EDTTYP);
               }
               auxdsc();
               locate(fields[ACTTYP].edtlocx,fields[ACTTYP].edtlocy);
               setatr(DSPATR);
               printf("%-6s",actyps[curchc]);
               dsphlp();
               break;
          case -F9:
               inloop=0;
               break;
          case -ALT_C:
               chscol();
               nsexploto(scntbl[MSCOBJ],0,0,7,5,59,1);
               proff(0,0);
               break;
          case -F10:
          case -TAB:
          case -BAKTAB:
          default:
               inloop=0;
          }
     } while (inloop);
     if (key == -ESC || key == -F9) {
          rc=ABNACT;
     }
     else {
          switch (key) {
          case -F10:
               rc=SAVACT;
               break;
          case -BAKTAB:
               rc=PRVFLD;
               break;
          case -TAB:
          default:
               rc=NXTFLD;
          }
          tmpinf.acttyp=actyps[curchc][0];
          fields[ACTTYP].contents=actyps[curchc];
          typidx=curchc;
     }
     popscn();
     dspfld(ACTTYP);
     auxdsc();
     chsngtyp=0;
     fstchc=sfst;
     curchc=scur;
     return(rc);
}

void
chkact(void)                         /* check action's flds at end of edit */
{
     int chkres;

     for (fnum=ACTNAM,edtnga=1 ; (fnum < NFLDS) && savit;
                                 fnum+=(fnum !=ACTNAM ? 1 : 2)) {
          if (fnum == ACTAUX && typidx != BOTH) {
               if (typidx == SIMPLE || typidx == DIRECT) {
                    tmpinf.actaux[0]='\0';
               }
               else {
                    chkbrc();
               }
          }
          else {
               while ((chkres=chkfld(fnum)) != FLDOK) {
                    dspfld(fnum);
                    dsphlp();
                    edterr(fnum,chkres);
                    if (edtfld(fnum) == ABNACT) {
                         savit=0;
                         break;
                    }
               }
               dspfld(fnum);
          }
     }
     edtnga=0;
}

void
chkbrc(void)                         /* check 4 "<" & ">" in "Aux Message" */
                                     /* for "Yell" type action             */
{
     int chkres,slen;

     ckpsax=0;
     if ((chkres=chkfld(ACTAUX)) != FLDOK) {
          dspfld(ACTAUX);
          slen=strlen(tmpinf.actaux);
          switch (chkres) {
          case NOLEFB:
               if (mssage("\"Tag Before Message\" has no LEFT bracket - "
                          "add it",1,"Yes",1)) {
                    if (slen < RSPLEN) {
                         movmem(tmpinf.actaux,tmpinf.actaux+1,
                                strlen(tmpinf.actaux)+1);
                    }
                    tmpinf.actaux[0]='<';
                    dspfld(ACTAUX);
               }
               break;
          case NORITB:
               if (mssage("\"Tag Before Message\" has no RIGHT bracket - "
                          "add it",1,"Yes",1)) {
                    if (slen < RSPLEN) {
                         tmpinf.actaux[slen+1]='\0';
                         tmpinf.actaux[slen]='>';
                    }
                    else {
                         tmpinf.actaux[slen-1]='>';
                    }
                    dspfld(ACTAUX);
               }
               break;
          case NOBRAK:
               if (mssage("\"Tag Before Message\" has NO brackets - add them",
                          1,"Yes",1)) {
                    if (slen < RSPLEN-1) {
                         movmem(tmpinf.actaux,tmpinf.actaux+1,slen+1);
                         tmpinf.actaux[0]='<';
                         slen++;
                         tmpinf.actaux[slen+1]='\0';
                         tmpinf.actaux[slen]='>';
                    }
                    else if (slen == RSPLEN-1) {
                         movmem(tmpinf.actaux,tmpinf.actaux+1,
                                strlen(tmpinf.actaux)+1);
                         tmpinf.actaux[0]='<';
                         tmpinf.actaux[slen-1]='>';
                    }
                    else {
                         tmpinf.actaux[0]='<';
                         tmpinf.actaux[slen-1]='>';
                    }
                    dspfld(ACTAUX);
               }
          }
     }
     ckpsax=1;
}

int
chkfld(                              /* check a field's correctness        */
int fnum)                            /*      field to check                */
{
     int rc=FLDOK;
     char *stg;

     stg=fields[fnum].contents;
     switch (fnum) {
     case ACTNAM:
          if (strlen(stg) == 0) {
               rc=NONAME;
          }
          else if (!sameas(orginf.actnam,stg)) {
               if (fndpos(stg) == NMEXST) {
                    rc=NMEXST;
               }
          }
          break;
     case ACTRES:
          if (strlen(stg) == 0) {
               rc=NOLEN;
          }
          break;
     case ACTOTH:
          if (strlen(stg) == 0) {
               rc=NOLEN;
          }
          else {
               rc=chkp_s(ACTOTH);
          }
          break;
     case ACTAUX:
          if (strlen(stg) == 0) {
               rc=NOLEN;
          }
          else {
               if (ckpsax) {
                    rc=chkp_s(ACTAUX);
               }
               else if (stg[0] != '<' && stg[strlen(stg)-1] != '>') {
                    rc=NOBRAK;
               }
               else if (stg[0] != '<') {
                    rc=NOLEFB;
               }
               else if (stg[strlen(stg)-1] != '>') {
                    rc=NORITB;
               }
          }
     }
     return(rc);
}

int
chkp_s (                             /* scan string and evaluate "%s"      */
int fnum)                            /*      field to scan                 */
{
     int rc=FLDOK,i=0,nump_s=0;

     switch (fnum) {
     case ACTOTH:
          if (typidx == SIMPLE || typidx == YELL) {
               minp_s=1;
               maxp_s=2;
          }
          else {
               minp_s=2;
               maxp_s=3;
          }
          break;
     case ACTAUX:
          if (typidx == BOTH) {
               minp_s=1;
               maxp_s=2;
          }
          else {
               minp_s=0;
               maxp_s=0;
          }
     }
     do {
          if (fields[fnum].contents[i] == '%') {
               switch (fields[fnum].contents[i+1]) {
               case 's':
                    if (nump_s < maxp_s) {
                         nump_s++;
                         i+=2;
                    }
                    else {
                         rc=TMNYPS;
                    }
                    break;
               case '%':
                    i+=2;
                    break;
               default:
                    rc=ILGCST;
               }
          }
          else {
               i++;
          }
     } while (rc == FLDOK && fields[fnum].contents[i] != '\0');
     if (nump_s < minp_s && rc != ILGCST) {
          rc=TFEWPS;
     }
     return(rc);
}

void
edterr(                              /* display appropriate error msgs     */
int fnum,                            /*      field containing error        */
int error)                           /*      error encountered             */
{
     dspnam(fnum,MSSATR+0x80);
     switch (error) {
     case TFEWPS:
          mssage(spr(flderr[TFEWPS],minp_s,(minp_s > 1 ? "s" : "")),0,"",1);
          break;
     case TMNYPS:
          if (minp_s > 0) {
               mssage(spr(flderr[TMNYPS],spr("more than %d",maxp_s)),
                      0,"",1);
          }
          else {
               mssage(spr(flderr[TMNYPS],"any"),0,"",1);
          }
          break;
     case ILGCST:
          mssage(spr("%s",flderr[ILGCST]),0,"",1);
          break;
     case NONAME:
          mssage(flderr[NONAME],0,"",1);
          break;
     case NOLEN:
          mssage(flderr[NOLEN],0,"",1);
          break;
     case NMEXST:
          dspfld(ACTNAM);
          mssage(spr("An action named \"%s\" already exists...",
                     tmpinf.actnam),0,"",1);
     }
     dspnam(fnum,NAMATR);
}

void
savact(                              /* save action in list                */
int actnum,                          /*     spot to save action in         */
int status)                          /*     new or existing?               */
{
     int pos;
     register int i;

     if (typidx == SIMPLE || typidx == DIRECT) {
          tmpinf.actaux[0]='\0';
     }
     if (status == NEW) {
          if ((pos=fndpos(tmpinf.actnam)) < nactns) {
               movmem(&anmlst[pos],&anmlst[pos+1],
                      sizeof(char *)*(nactns-pos));
          }
          anmlst[pos]=alczer(NAMLEN+1);
          stzcpy(anmlst[pos],tmpinf.actnam,NAMLEN+1);
          if (nactns > AINWIN) {
               fstchc=(pos < nactns-AINWIN ? pos : nactns-AINWIN+1);
          }
          else {
               fstchc=0;
          }
          curchc=pos;
          chgptrs(actnum,(pos < nactns ? pos : EOLST),status);
     }
     else if (!sameas(orginf.actnam,tmpinf.actnam)) {
          if (sameto(tmpinf.actnam,orginf.actnam)
              || sameto(orginf.actnam,tmpinf.actnam)
              || (pos=fndpos(tmpinf.actnam)) == actnum
              || pos == actnum+1) {
               stzcpy(anmlst[actnum],tmpinf.actnam,NAMLEN+1);
          }
          else {
               if (pos < actnum) {
                    for (i=actnum ; i > pos ; i--) {
                         stzcpy(anmlst[i],anmlst[i-1],NAMLEN+1);
                    }
                    stzcpy(anmlst[pos],tmpinf.actnam,NAMLEN+1);
                    fstchc=(pos < nactns-AINWIN ? pos : nactns-AINWIN);
                    curchc=pos;
               }
               else {
                    for (i=actnum ; i < pos-1 ; i++) {
                         stzcpy(anmlst[i],anmlst[i+1],NAMLEN+1);
                    }
                    stzcpy(anmlst[pos-1],tmpinf.actnam,NAMLEN+1);
                    fstchc=(pos-1 < nactns-AINWIN ? pos-1 : nactns-AINWIN);
                    curchc=pos-1;
               }
               chgptrs(actnum,(pos < nactns ? pos : EOLST),status);
          }
     }
     tmp->type=tmpinf.acttyp;
     for (i=ACTNAM ; i < NFLDS ; i++) {
          if (i != ACTTYP) {
               if (status == NEW
                  || strcmp(fields[i].contents,fields[i].ocontnts) != 0) {
                    savfld(fields[i].contents,i,fields[i].edtlen);
               }
          }
     }
     tmp->color=tmpinf.color;
     chgsmd=1;
}

void
chgptrs(                             /* change ptrs in action list to add, */
                                     /* change pos, delete                 */
int actnum,                          /*      current position of action    */
int pos,                             /*      new position of action        */
int status)                          /*      new, existing, or delete?     */
{
     struct action *pbpos,*pbanm,*ppos;

     switch (status) {
     case NEW:
          switch (pos) {
          case 0:
               tmp->nxtact=actlst;
               actlst=tmp;
               break;
          case EOLST:
               aclstl=aclstl->nxtact=tmp;
               tmp->nxtact=NULL;
               break;
          default:
               pbpos=fndact(pos-1);
               tmp->nxtact=pbpos->nxtact;
               pbpos->nxtact=tmp;
          }
          break;
     case EXISTS:
          switch (pos) {
          case 0:
               pbanm=fndact(actnum-1);
               pbanm->nxtact=tmp->nxtact;
               tmp->nxtact=actlst;
               actlst=tmp;
               break;
          case EOLST:
               if (actnum != 0) {
                    pbanm=fndact(actnum-1);
                    pbanm->nxtact=tmp->nxtact;
               }
               else {
                    actlst=actlst->nxtact;
               }
               tmp->nxtact=NULL;
               aclstl=aclstl->nxtact=tmp;
               break;
          default:
               pbpos=fndact(pos-1);
               pbanm=fndact(actnum-1);
               ppos=fndact(pos);
               pbanm->nxtact=tmp->nxtact;
               pbpos->nxtact=tmp;
               tmp->nxtact=ppos;
          }
          break;
     case DELETE:
          if (actnum == 0) {
               actlst=actlst->nxtact;
          }
          else {
               (pbanm=fndact(actnum-1))->nxtact=(actnum != nactns-1
                                                 ? tmp->nxtact
                                                 : NULL);
          }
          free(tmp);
     }
}

void
savfld(                              /* alloc mem / save contents of field */
char *contents,                      /*      field contents                */
int fnum,                            /*      field currently being edited  */
int len)                             /*      field len (incl. '\0')        */
{
     if (tmp->chginf == NULL) {
          tmp->chginf=(char **)alczer(NFLDS*(sizeof(char *)));
          tmp->chginf[ACTTYP]=NULL;
     }
     if (tmp->chginf[fnum] == NULL) {
          tmp->chginf[fnum]=alczer(len);
     }
     stzcpy(tmp->chginf[fnum],contents,len);
}

int
memlow(void)                         /* Warn if it's not safe to continue  */
{
     int rc=0;

     if (sizmem() < MINMEM) {
          mssage("Memory low!  Save list (hit ALT-S) before continuing...",
                 0,"",1);
          rc=1;
     }
     return(rc);
}

void
auxdsc(void)                         /* dynamically remove/rename          */
                                     /* "Auxiliary Message" field          */
{
     char buffer[DSPLEN+1];

     setatr(!edtnga || fnum != ACTAUX ? NAMATR : CRNATR);
     locate(fields[ACTAUX].nmlocx,fields[ACTAUX].nmlocy);
     switch (typidx) {
     case SIMPLE:
     case DIRECT:
          fields[ACTAUX].name="";
          break;
     case BOTH:
          fields[ACTAUX].name="Auxiliary Message:";
          break;
     case YELL:
          fields[ACTAUX].name="Tag Before Message:";
     }
     printf("%-19s",fields[ACTAUX].name);
     if (chsngtyp) {
           switch (typidx) {
           case SIMPLE:
           case DIRECT:
                setatr(DIMATR);
                break;
           case BOTH:
                setatr(tmpinf.color);
                break;
           case YELL:
                setatr(DSPATR);
           }
           locate (fields[ACTAUX].edtlocx,fields[ACTAUX].edtlocy);
           if (strlen(fields[ACTAUX].contents) > DSPLEN) {
                stzcpy(buffer,fields[ACTAUX].contents,DSPLEN+1);
                printf("%-65s",buffer);
           }
           else {
                printf("%s",spr("%-*s",strlen(fields[ACTAUX].contents),
                                fields[ACTAUX].contents));
           }
     }
}

void
dspact(void)                         /* display all fields for an action   */
{
     register int i;

     for (i=0 ; i < NFLDS ; i++) {
          dspfld(i);
     }
}

void
dspfld(                              /* display contents of a field        */
int fnum)                            /*      field to display              */
{
     char buffer[DSPLEN+1];

     locate(fields[fnum].edtlocx,fields[fnum].edtlocy);
     setatr(BLKATR);
     printf("%s",spr("%-*s",fields[fnum].blklen,""));
     locate(fields[fnum].edtlocx,fields[fnum].edtlocy);
     setatr(fldclr(fnum,1));
     if (strlen(fields[fnum].contents) > DSPLEN) {
          stzcpy(buffer,fields[fnum].contents,DSPLEN+1);
          printf("%-65s",buffer);
     }
     else {
          printf("%s",spr("%-*s",strlen(fields[fnum].contents),
                          fields[fnum].contents));
     }
}

void
dsphlp(void)                         /* display appropriate help msg's     */
{
      char *hlpfil;
      int upy;

      if (!edtnga && !chsngtyp) {
           hlpfil=scntbl[PTTYPH];
           upy=0;
      }
      else {
           switch (fnum) {
           case ACTNAM:
           case ACTRES:
           case ACTUID:
                hlpfil=scntbl[fields[fnum].hlpfil];
                upy=fields[fnum].hlpupy;
                break;
           case ACTTYP:
                hlpfil=scntbl[PTTYPH];
                upy=typhlp[typidx];
                break;
           case ACTOTH:
                switch (typidx) {
                case SIMPLE:
                case YELL:
                     hlpfil=scntbl[fields[ACTOTH].hlpfil];
                     upy=fields[ACTOTH].hlpupy;
                     break;
                case DIRECT:
                case BOTH:
                     hlpfil=scntbl[fields[ACTOTH].ahlpfil];
                     upy=fields[ACTOTH].ahlpupy;
                }
                break;
           case ACTAUX:
                switch (typidx) {
                case BOTH:
                     hlpfil=scntbl[fields[ACTAUX].hlpfil];
                     upy=fields[ACTAUX].hlpupy;
                     break;
                case YELL:
                     hlpfil=scntbl[fields[ACTAUX].ahlpfil];
                     upy=fields[ACTAUX].ahlpupy;
                }
           }
      }
      nsexploto(hlpfil,HLPUPX,upy,HLPUPX+HLNLEN-1,upy+NHLPLN-1,13,15);
      proff(0,0);
}

void
dspnac(void)                         /* display # of actions               */
{
     locate (74,0);
     setatr(0x70);
     printf("%3d",nactns);
}

void
dspnam(                              /* display name of a field            */
int fldnum,                          /*      field's name to display       */
int attr)                            /*      way in which to display it    */
{
     locate(fields[fldnum].nmlocx,fields[fldnum].nmlocy);
     setatr(attr);
     printf("%s",fields[fldnum].name);
}

int
fldclr(                              /* set field edit/display color       */
int fnum,                            /*      field for which to set color  */
int st4dsp)                          /*      1: display clr  0: edit clr   */
{
     int rc;

     switch (fnum) {
     case ACTNAM:
     case ACTTYP:
     case ACTUID:
          rc=(st4dsp ? DSPATR : NCEATR);
          break;
     case ACTRES:
     case ACTOTH:
          rc=tmpinf.color;
          break;
     case ACTAUX:
          if (typidx == YELL) {
               rc=(st4dsp ? DSPATR : NCEATR);
          }
          else {
               rc=tmpinf.color;
          }
     }
     return(rc);
}

int
fndpos (                             /* find where to insert action        */
char *nam)                           /*      action's name                 */
{
     int low,mid,hi,resa,resb,rc;

     if (strcmp(nam,anmlst[0]) < 0) {
          rc=0;
     }
     else if (strcmp(nam,anmlst[nactns-1]) > 0) {
          rc=nactns;
     }
     else {
          low=0;
          hi=nactns-1;
          while (1) {
               mid=(low+hi)/2;
               resa=strcmp(nam,anmlst[mid]);
               resb=strcmp(nam,anmlst[mid-1]);
               if (resa < 0 && resb > 0) {
                    rc=mid;
                    break;
               }
               else if (resa == 0 || resb == 0) {
                    rc=NMEXST;
                    break;
               }
               else if (resa < 0) {
                    hi=mid-1;
               }
               else {
                    low=mid+1;
               }
          }
     }
     return(rc);
}

struct action *
fndact(                              /* find action at spec pos in list    */
int actnum)                          /*      pos of action to search for   */
{
     struct action *tp;

     register int i;

     for (i=0,tp=actlst ; i < actnum ; i++,tp=tp->nxtact) {
     }
     return(tp);
}

void
indchg(void)                         /* indicate if field is changed       */
{
     switch (edtvalc) {
     case HOME:
     case END:
     case PGUP:
     case PGDN:
          break;
     case '\b':
          if (edtbof == 0 && edtxof == 0) {
               break;
          }
     case ' ':
          if (fnum == ACTNAM) {
               break;
          }
     default:
          fldchg=1;
          keyhlp(EDTFLD+RSFHLP);
     }
}

void
keyhlp(                              /* show appropriate key help          */
int n)                               /*      key help to show              */
{
     nsexploto(scntbl[MSCOBJ],14,n,78,n+2,13,21);
     proff(0,0);
}

int
kyhdlr(                              /* key validation rtn for edtval()    */
int c,                               /*      key pressed                   */
char *sval)                          /*      buffer contents so far        */
{
     int rc;

     if (ynvalid) {
          rc=0;
          if (toupper(c) == 'Y') {
               stzcpy(sval,"Yes",4);
          }
          else if (toupper(c) == 'N') {
               stzcpy(sval,"No ",4);
          }
     }
     else if (chsngcol) {
          rc=(c >= '0' && c <= '9');
     }
     else {
          if (!isascii(c) || c == '\n' || c == '\f' || c == '\7') {
               rc=0;
          }
          else {
               switch(fnum) {
               case ACTNAM:
                    rc=isalpha(c);
                    break;
               case ACTUID:
                    if (isdigit(c)) {
                         rc=digalw;
                    }
                    else if (strchr(PUNCHR,c) != NULL) {
                         rc=fulalw;
                    }
                    else {
                         rc=isuidc(c);
                    }
                    break;
               case ACTRES:
                    rc=(c != '%');
                    break;
               case ACTAUX:
                    rc=(typidx != YELL ? 1 : c != '%');
                    break;
               default:
                    rc=1;
               }
          }
     }
     return(rc);
}

void
ldtmpa(                              /* load tmpinf with action info, or   */
                                     /* all blank if adding action         */
int actnum)                          /*      action# to load, or NEW/CURACT*/
{
     register int i,ndaux=0;

     switch (actnum) {
     default:
          tmp=fndact(actnum);
          ndaux=1;
     case CURACT:
          tmpinf.acttyp=tmp->type;
          sttidx();
          for (i=ACTNAM ; i < NFLDS ; i++) {
               if (i != ACTTYP) {
                    getfld(fields[i].contents,
                           tmp->chginf == NULL ? NULL : tmp->chginf[i],
                           i,fields[i].edtlen);
               }
          }
          tmpinf.color=tmp->color;
          break;
     case NEWACT:
          setmem(&tmpinf,sizeof(struct actinf),'\0');
          tmpinf.acttyp='S';
          tmpinf.color=curclr;
          typidx=SIMPLE;
          ndaux=1;
          break;
     }
     fields[ACTTYP].contents=actyps[typidx];
     if (ndaux) {
          auxdsc();
     }
}

void
getfld(                              /* get string for field edit          */
char *dest,                          /*      where to put the string       */
char *src,                           /*      indicate where to get string  */
int fnum,                            /*      field currently being edited  */
int len)                             /*      field length (incl. '\0')     */
{
     char mcvbuf[RSPLEN+LONGCL+1];

     if (src == NULL) {
          if (tmp->msgnum != NONE) {
               stzcpy(mcvbuf,rawmsg(tmp->msgnum+fnum),RSPLEN+LONGCL+1);
               stpans(mcvbuf);
               stzcpy(dest,mcvbuf,len);
          }
          else {
               dest[0]='\0';
          }
     }
     else {
          stzcpy(dest,src,len);
     }
}

int
mssage(                              /* display various messages to user   */
char *mlin,                          /*      message to display            */
int  ynmsg,                          /*      wait for yes/no response?     */
char *yndef,                         /*      default for yes/no response   */
int  pause)                          /*      pause, wait for keypress?     */
{
     int rc=1,offset,mlen,ynpos;
     char ynans[4];

     pushscn();
     nsexploto(scntbl[MSCOBJ],0,12,64,14,13,21);
     proff(0,0);
     setatr(MSSATR);
     if (!ynmsg) {
          locate(14+(63-strlen(mlin))/2,22);
          printf("%-s",mlin);
          if (pause) {
               locate(37,23);
               printf("(Hit any key...)");
               getchc();
               popscn();
          }
     }
     else {
          ynvalid=1;
          offset=10+(63-(mlen=strlen(mlin)))/2;
          ynpos=offset+mlen+3;
          locate(offset,22);
          printf("%-s? ",mlin);
          stzcpy(ynans,yndef,4);
          if (!edtval(ynpos,22,4,ynans,kyhdlr,MCHOICE)
              || ynans[0] == 'N') {
               rc=0;
          }
          ynvalid=0;
          popscn();
     }
     return(rc);
}

int
mychoo(                              /* select a choice / choose function  */
int n,                               /*      number of choices             */
char *chcs[],                        /*      text of choices               */
int upx,                             /*      upper-x \                     */
int upy,                             /*      upper-y  \ screen coords of   */
int lox,                             /*      lower-x  /    choose window   */
int loy,                             /*      lower-y /                     */
int escok)                           /*      escape key ok?                */
{
     int c,rc,sfst,scur;

     if (fstchc >= nactns-AINWIN) {
          movmem(scntbl[0]+scnoff(3,24),frzseg()+scnoff(3,24),12);
     }
     sfst=fstchc;
     scur=curchc;
     supchc(n,chcs,upx,upy,lox,loy,escok);
     fstchc=sfst;
     curchc=scur;
     dspchc();
     do {
          switch (c=getchc()) {
          case CRSRUP:
          case CRSRDN:
          case PGUP:
          case PGDN:
               if (c == CRSRUP || c == PGUP ? curchc > 0 : curchc < n-1) {
                    rc=REDISP;
                    hdlchc(c);
               }
               else {
                    rc=n;
               }
               break;
          case TAB:
          case BAKTAB:
               rc=(chsngtyp ? -c : n);
               break;
          case INS:
          case DEL:
          case F10:
          case F9:
          case ESC:
          case ALT_S:
          case ALT_C:
               rc=-c;
               break;
          default:
               if (isascii(c) && isalpha(c)) {
                    hdlchc(c);
                    if (chsngtyp) {
                         rc=REDISP;
                    }
                    else {
                         ldtmpa(curchc);
                         dspact();
                         rc=n;
                    }
               }
               else {
                    rc=hdlchc(c);
               }
          }
     } while (rc == n);
     return(rc);
}

void
pushscn(void)                        /* save current screen                */
{
     if (screens != NULL) {
          curscn=curscn->nxtscn=
                            (struct scninfo *)alcmem(sizeof(struct scninfo));
     }
     else {
          curscn=screens=(struct scninfo *)alcmem(sizeof(struct scninfo));
     }
     scn2mem(frzseg(),curscn->scn=alcmem(SCNSIZ),SCNSIZ);
     curscn->color=(edtnga ? tmpinf.color : tmp->color);
     curscn->nxtscn=NULL;
}

void
popscn(void)                         /* restore last saved screen          */
{
     struct scninfo *newcur;
     int oclr;
     register int i;

     if (screens != NULL) {
          if (curscn != screens) {
               movmem(curscn->scn,frzseg(),SCNSIZ);
               free(curscn->scn);
               for (newcur=screens ; newcur->nxtscn->nxtscn != NULL
                    ; newcur=newcur->nxtscn) {
               }
               newcur->nxtscn=NULL;
               oclr=curscn->color;
               free(curscn);
               curscn=newcur;
               if (oclr != tmpinf.color) {
                    for (i=ACTRES ; i <= ACTAUX ; i++) {
                         dspfld(i);
                    }
               }
          }
          else {
               movmem(curscn->scn,frzseg(),SCNSIZ);
               free(curscn->scn);
               free(curscn);
               curscn=screens=NULL;
          }
     }
     else {
          catastro("popscn: screen stack is empty...");
     }
}

void
rdacts(void)                         /* read action list from GALETL.MCV   */
{
     register int i,cmsgnm;
     char buffer[RSPLEN+LONGCL+1];
     int clrtyp;
     FILE *hfile;

     mssage("Reading actions...",0,"",0);
     galetl=opnmsg("GALETL.MCV");
     setmbk(galetl);
     if ((hfile=fopen("GALETL.H",FOPRA)) == NULL) {
          catastro("CANNOT FIND \"GALETL.H\"");
     }
     nactns=numopt((nacmsg=msgnum(hfile,"NACTNS")),4,600);
     fclose(hfile);
     anmlst=(char **)alczer(nactns*sizeof(char *));
     tmp=actlst=(struct action *)alczer(sizeof(struct action));
     for(i=0,cmsgnm=nacmsg+2 ; i < nactns ; i++,cmsgnm+=NFLDS) {
          tmp->type=chropt((tmp->msgnum=cmsgnm)+ACTTYP);
          tmp->chginf=NULL;
          stzcpy(buffer,rawmsg(cmsgnm+ACTRES),LONGCL+1);
          if ((clrtyp=bgclsq(buffer)) == NOCL) {
               tmp->color=curclr;
          }
          else {
               tmp->color=str2atr(buffer,clrtyp);
          }
          anmlst[i]=alczer(NAMLEN+1);
          stzcpy(buffer,rawmsg(cmsgnm),NAMLEN+1);
          stpans(buffer);
          stzcpy(anmlst[i],buffer,NAMLEN+1);
          if (i < nactns-1) {
               tmp->nxtact=(struct action *)alcmem(sizeof(struct action));
               tmp=tmp->nxtact;
          }
     }
     (aclstl=tmp)->nxtact=NULL;
     popscn();
     dspnac();
}

int
bgclsq(       /* Use "brute force" to determine if an action begins with   */
              /* an ANSI color sequence - must be in one of these forms:   */
              /* <esc> [ <0/1/5> ; 3 <0-7> ; 4 <0-7> m                     */
              /* <esc> [ <0/1/5> ; <0/1/5> ; 3 <0-7> ; 4 <0-7> m           */
              /* <esc> [ <0/1/5> ; <0/1/5> ; <0/1/5> ; 3 <0-7> ; 4 <0-7> m */
char *seq)    /*      sequence to check                                    */
{
     int rc=NOCL;

     if (seq[0] == 27 && seq[1] == '['
         && (seq[2] == '0' || seq[2] == '1' || seq[2] == '5')
         && seq[3] == ';') {
          if (seq[4] == '3' && seq[5] >= '0' && seq[5] <= '7'
              && seq[6] == ';'
              && seq[7] == '4' && seq[8] >= '0' && seq[8] <= '7'
              && seq[9] == 'm') {
               rc=SHRTCL;
          }
          else if ((seq[4] == '0' || seq[4] == '1' || seq[4] == '5')
                   && seq[5] == ';'
                   && seq[6] == '3' && seq[7] >= '0' && seq[7] <= '7'
                   && seq[8] == ';'
                   && seq[9] == '4' && seq[10] >= '0' && seq[10] <= '7'
                   && seq[11] == 'm') {
               rc=MIDLCL;
          }
          else if ((seq[4] == '0' || seq[4] == '1' || seq[4] == '5')
                   && seq[5] == ';'
                   && (seq[6] == '0' || seq[6] == '1' || seq[6] == '5')
                   && seq[7] == ';'
                   && seq[8] == '3' && seq[9] >= '0' && seq[9] <= '7'
                   && seq[10] == ';'
                   && seq[11] == '4' && seq[12] >= '0' && seq[12] <= '7'
                   && seq[13] == 'm') {
               rc=LONGCL;
          }
     }
     return(rc);
}

void
rdfile(                              /* read lines from/wrt to ascii file  */
                                     /* until spec. string                 */
FILE *frmfle,                        /*      file to read from             */
FILE *tofle,                         /*      file 2 write 2 (NULL: no wrt) */
char *endstr,                        /*      string to stop at             */
int  pos,                            /*      pos want string in (or ANYPOS)*/
int wrends)                          /*      write line w/ ending string?  */
{
     char *stgpos;

     while (1) {
          if (fgets(line,MXBFSZ,frmfle) == NULL) {
               catastro("RDFILE: Cannot find string \"%s\" at column %d",
                        endstr,pos);
          }
          stgpos=strstr(line,endstr);
          if (pos == ANYPOS ? stgpos != NULL : stgpos == line+pos) {
               break;
          }
          else {
               putline(tofle);
          }
     }
     if (wrends) {
          putline(tofle);
     }
}

void
putline(                             /* write line[] buffer to file, if any */
FILE *fp)                            /*   file to write to, or NULL if none */
{
     if (fp != NULL) {
          fprintf(fp,"%s",line);
     }
}

void
rszlst(                              /* re-alloc less/more mem 4 name list */
int newsize)
{
     anmlst=(char **)alcrsz(anmlst,nactns*sizeof(char *),
                            newsize*sizeof(char *));
}

void
sttidx(void)                         /* indicate type of action            */
{
     switch (tmpinf.acttyp) {
     case 'S':
          typidx=SIMPLE;
          break;
     case 'D':
          typidx=DIRECT;
          break;
     case 'B':
          typidx=BOTH;
          break;
     case 'Y':
          typidx=YELL;
          break;
     default:
          catastro(spr("sttidx: bad action type: %c/action: %s/%d",
                       tmpinf.acttyp,tmpinf.actnam,tmp->msgnum));
     }
}

int                                  /*     1:exit  0:ESC hit, don't exit  */
wracts(                              /* wrte action list to new GALETL.MSG */
                                     /* & create new GALETL.MCV & GALETL.H */
int exitng)                          /*     being called just before exit? */
{
     register int i,j,actwtn;
     int uaclst,rc=1;
     char nmsufx[4];
     FILE *galetlm,*tmpfle;

     if ((tmpfle=fopen("GALETL.NEW",FOPWA)) == NULL) {
          catastro("COULD NOT OPEN \"GALETL.NEW\"!");
     }
     if ((galetlm=fopen("GALETL.MSG",FOPRA)) == NULL) {
          catastro("COULD NOT OPEN \"GALETL.MSG\"!");
     }
     nsexploto(scntbl[COLORB],HLPUPX,18,HLPUPX+HLNLEN-1,18+NHLPLN-1,13,15);
     uaclst=mssage("Update \"ACTLST\"",1,"Yes",1);
     if (edtvalc != ESC) {
          if (!exitng) {
               dsphlp();
          }
          mssage("Writing actions to \"GALETL.MSG\"...",0,"",0);
          if (uaclst) {
               rdfile(galetlm,tmpfle,"ACTLST {",0,0);
               fprintf(tmpfle,"ACTLST {\33[0m\n");
               fprintf(tmpfle,"          \33[1;32mAction Word Listing\n");
               fprintf(tmpfle,"          ===================\n\n");
               for (i=0,j=0,actwtn=0,tmp=actlst ;
                                    i < nactns && actwtn < ACTWLM;
                                                       i++,tmp=tmp->nxtact) {
                    ldtmpa(CURACT);
                    if (tmpinf.actuid[0] == '\0') {
                         fprintf(tmpfle,"%-13s",anmlst[i]);
                         actwtn++;
                         j++;
                         if (j%ACTPLN == 0) {
                              fprintf(tmpfle,"\n");
                         }
                    }
               }
               if (!(j%ACTPLN == 0)) {
                    fprintf(tmpfle,"\n");
               }
               fprintf(tmpfle,"} T Teleconference action word list header\n");
               rdfile(galetlm,NULL,"}",ANYPOS,0);
          }
          rdfile(galetlm,tmpfle,"NACTNS {",0,0);
          fprintf(tmpfle,"NACTNS {Number of action commands: %d} "
                         "N 1 600\n",nactns);
          rdfile(galetlm,tmpfle,"}",ANYPOS,1);
          fprintf(tmpfle,"\n");
          for(i=0,tmp=actlst ; i < nactns ; i++,tmp=tmp->nxtact) {
               stzcpy(nmsufx,spr("%03d",i+1),4);
               ldtmpa(CURACT);
               atr2str(tmp->color);
               fprintf(tmpfle,"ACT%s {%s} "
                       "S 11 DO NOT CHANGE \"ACTnnn\" IN CNF - USE GALAED - "
                       "SEE DOCS!\n",
                       nmsufx, tmpinf.actnam);
               fprintf(tmpfle,"TYP%s {%c} C Type of action\n",
                       nmsufx,tmpinf.acttyp);
               fprintf(tmpfle,"RES%s {%s%s} "
                       "T Teleconference response to user who issued action\n",
                       nmsufx,ansseq,tmpinf.actres);
               fprintf(tmpfle,"OTH%s {%s%s} "
                       "T Teleconference what other users see\n",
                       nmsufx,ansseq,tmpinf.actoth);
               if (tmpinf.acttyp == 'B') {
                    fprintf(tmpfle,"AUX%s {%s%s} "
                            "T Teleconference aux. msg. / tag before msg. "
                            "if necessary\n",
                            nmsufx,ansseq,tmpinf.actaux);
               }
               else {
                    fprintf(tmpfle,"AUX%s {%s} T Teleconference aux. msg. / "
                            "tag before msg. if necessary\n",
                            nmsufx,tmpinf.actaux);
               }
               fprintf(tmpfle,"UID%s {%s} "
                       "T Teleconference action reserved for this user "
                       "only\n\n",nmsufx,tmpinf.actuid);
          }
          popscn();
          fclose(tmpfle);
          fclose(galetlm);
          if (rename("GALETL.MSG","GALETL.OLD") != 0) {
               catastro("CANNOT RENAME GALETL.MSG TO GALETL.OLD!");
          }
          if (rename("GALETL.NEW","GALETL.MSG") != 0) {
               catastro("CANNOT RENAME GALETL.NEW TO GALETL.MSG!");
          }
          if (remove("GALETL.OLD") != 0) {
               catastro("CANNOT DELETE GALETL.OLD!");
          }
          clsmsg(galetl);
          system("bbsmsx galetl");
          cursiz(NOCURS);
     }
     else {
          rc=0;
          dsphlp();
     }
     return(rc);
}

void
atr2nms(                             /* hex attr # -> nums for chscol()    */
int attr,                            /*      "hex" attribute number        */
int *fore,                           /*      ptr to return foreground clr  */
int *back)                           /*      ptr to return background clr  */
{
     int blink=0;

     if (attr > 0x7F) {
          blink=16;
          attr-=0x80;
     }
     *back=attr>>4;
     *fore=attr-(*back<<4)+blink;
}

void
atr2str(                             /* hex attr # -> ANSI color sequence  */
int attr)                            /*      "hex" atrribute number        */
{
     int attrib,forepos,backpos,b_npos,foreval,backval;

     attrib=attr;
     if (attrib > 0x7F) {
          stzcpy(ansseq,"\33[0;5; ;3 ;4 m",15);
          attrib-=0x80;
          forepos=9;
          backpos=12;
          b_npos=6;
     }
     else {
          stzcpy(ansseq,"\33[0; ;3 ;4 m",13);
          forepos=7;
          backpos=10;
          b_npos=4;
     }
     backval=attrib>>4;
     foreval=attrib-(backval<<4);
     if (foreval > 7 && foreval < 16) {
          foreval-=8;
          ansseq[b_npos]='1';
     }
     else if (foreval > 23) {
          foreval-=16;
          ansseq[b_npos]='1';
     }
     else {
          ansseq[b_npos]='0';
     }
     ansseq[forepos]=asct[foreval]+48;
     ansseq[backpos]=asct[backval]+48;
}

int nms2atr(                         /* nums from chscol() -> hex attr #   */
int fore,                            /*      forground color number        */
int back)                            /*      background color number       */
{
     int foreval,backval,blink=0x00;

     foreval=fore;
     backval=back;
     if (foreval > 15) {
          foreval-=16;
          blink=0x80;
     }
     return(foreval+(backval<<4)+blink);
}

int
str2atr(                             /* ANSI color string -> hex attr #    */
char *seq,                           /*      ANSI sequence                 */
int seqtyp)                          /*      long, med, or short seq.      */
{
     int fore,back,forepos,backpos,blink=0x00,i,last;

     switch (seqtyp) {
     case SHRTCL:
          last=2;
          forepos=5;
          backpos=8;
          break;
     case MIDLCL:
          last=4;
          forepos=7;
          backpos=10;
          break;
     case LONGCL:
          last=6;
          forepos=9;
          backpos=12;
     }
     fore=asct[seq[forepos]-48];
     back=asct[seq[backpos]-48]<<4;
     for (i=2 ; i <= last ; i+=2) {
          switch (seq[i]) {
          case '0':
               fore-=(fore > 0x08 ? 0x08 : 0x00);
               break;
          case '1':
               fore+=(fore < 0x08 ? 0x08 : 0x00);
               break;
          case '5':
               blink+=(blink == 0x00 ? 0x80 : 0x00);
          }
     }
     return(fore+back+blink);
}
