/***************************************************************************
 *                                                                         *
 *   GALSEF.C      (Special Edit Functions)                                *
 *                                                                         *
 *   Copyright (C) 1992 GALACTICOMM, Inc.    All Rights Reserved.          *
 *                                                                         *
 *   This utility, based on the GCOMM.LIB routine EDTVAL.C, is used by     *
 *   GALAED.C for editing data fields of arbitrary types.  Like EDTVAL.C,  *
 *   it has full support for cursor movement, ins/del, home/end, etc.  The *
 *   global character edtvalc may be used upon exit to determine where to  *
 *   go next (its value can be '\r', TAB, BAKTAB, CRSRUP, CRSRDN, or ESC). *
 *   If the function is used in "exploded" form, the global variable       *
 *   edtedw may be used for a horizontal scroll, and the global variable   *
 *   edtxof may be used to control where the cursor will initially be      *
 *   located in the edit window at the start of edit.  If the ESC key is   *
 *   pressed, edtval() returns 0, otherwise it returns 1.                  *
 *                                                                         *
 *                                     - Bill Hyatt & R. N. Stein          *
 *                                       12/2/92                           *
 *                                                                         *
 ***************************************************************************/

#include "gcomm.h"
#include "galsef.h"

#define MXWNSZ 80             /* assumed maximum window size               */
#define MXBFSZ 256            /* assumed maximum buffer size               */

int edtvalc;                  /* last char accepted by edtval()            */
int edtxof;                   /* cursor X offset from start of field       */
int edtedw;                   /* size of edit field when scrolling         */
int edtbof;                   /* offset from beg of buffer when scrolling  */

extern far
int pxoff,pyoff;              /* proff() offsets used by edtval() & prat() */

STATIC int gorite(void);

static int inson=0;           /* insert on or off?                         */
static int savatr,savecur;    /* screen attr, cursor size before edit      */
static int savexo,saveyo;     /* save x,y offsets in key hdlr routines     */
static char svsbuf[MXBFSZ+1]; /* buffer contents before edit               */
static int mxlnm1;            /* field length, w/out '\0'                  */
static int strtx,strty;       /* x,y edit coord, incl. proff() offsets     */
static char *edsbuf;          /* buffer being edited                       */
static int (*validrou)();     /* validation routine address                */
static int flgset;            /* option flags (see gcomm.h)                */
static int strpl;             /* util. var. - strpln(edsbuf)               */

int                                     /*   0 if ESC hit; otherwise 1     */
edtval(sx,sy,maxlen,sval,valrou,flags)  /* edit a string value field       */
int sx,sy;                              /*   starting x/y coords (0-based) */
int maxlen;                             /*   max len of field (incl. '\0') */
char *sval;                             /*   value edit and return-buffer  */
int (*valrou)();                        /*   validation routine address    */
int flags;                              /*   option flags (see gcomm.h)    */
{
     supedt(sx,sy,maxlen,sval,valrou,flags);
     do {
          rfredt(color ? 0x0F : 0x70);
          edtvalc=getchc();
     } while(hdledt());
     endedt();
     return(edtvalc != ESC);
}

void
supedt(sx,sy,maxlen,sval,valrou,flags)  /* set up to edit a string          */
int sx,sy;                              /*   starting x/y coords (0-based)  */
int maxlen;                             /*   max length of field, incl '\0' */
char *sval;                             /*   value edit and return-buffer   */
int (*valrou)();                        /*   validation routine address     */
int flags;                              /*   option flags (see gcomm.h)     */
{
     strtx=sx;
     strty=sy;
     if ((flgset=flags)&USEPOFF) {
          strtx+=pxoff;
          strty+=pyoff;
     }
     mxlnm1=edtedw=maxlen-1;
     stzcpy(svsbuf,edsbuf=sval,sizeof(svsbuf));
     validrou=valrou;
     savatr=*(frzseg()+scnoff(strtx,strty)+1);
     savecur=curcurs();
     cursiz(LILCURS);
     padfld(edsbuf,maxlen);
     edtxof=((strpl=strpln(edsbuf)) < mxlnm1 ? strpl : strpl-1);
     edtbof=0;
}

void
rfredt(atr)                             /* refresh display of edited field */
int atr;                                /*    attribute for display        */
{
     char dspbuf[MXWNSZ+1];
     int rfrsvc;

     rfrsvc=curcurs();
     cursiz(NOCURS);
     locate(strtx,strty);
     setatr(atr);
     if (edtbof > mxlnm1-edtedw) {
          edtbof=mxlnm1-edtedw;
     }
     stzcpy(dspbuf,edsbuf+edtbof,edtedw+1);
     printf("%s",spr("%-*s",edtedw,dspbuf));
     if (!(flgset&MCHOICE)) {
          if (edtxof > edtedw-1) {
               edtxof=edtedw-1;
          }
          cursiz(rfrsvc);
          locate(strtx+edtxof,strty);
     }
}

int                                     /*   1: keep editing  0: end editing */
hdledt(void)                            /* hdl key in edit session           */
{                                       /*    implicit input: edtvalc        */
     int oedtbof,rc=1;

     switch (edtvalc) {
     case ESC:
          strcpy(edsbuf,svsbuf);
          rc=0;
          break;
     case TAB:
     case BAKTAB:
     case CRSRDN:
     case CRSRUP:
          if (!(flgset&MULTIEX)) {
               break;
          }
     case '\r':
          rc=0;
          break;
     case '\b':
          if ((edtbof == 0 && edtxof == 0) || flgset&MCHOICE) {
               break;
          }
          else if (edtxof > 0) {
               edtxof--;
          }
          else {
               oedtbof=edtbof;
               edtbof=(edtbof > edtedw ? edtbof-edtedw : 0);
               edtxof=oedtbof-edtbof-1;
          }
     case DEL:
          if (!(flgset&MCHOICE)) {
               movmem(edsbuf+edtbof+edtxof+1,edsbuf+edtbof+edtxof,
                      strlen(edsbuf+edtbof+edtxof));
               edsbuf[strlen(edsbuf)]=' ';
          }
          break;
     case CRSRLF:
          if (!(flgset&MCHOICE)) {
               if (edtxof > 0) {
                    edtxof--;
                    break;
               }
               else if (edtbof > 0) {
                    edtbof--;
                    break;
               }
          }
          if (flgset&MULTIEX) {
               rc=0;
          }
          break;
     case CRSRRT:
          if (!(flgset&MCHOICE)
             && (edsbuf[edtbof+edtxof] != ' '
                || edtbof+edtxof < strpln(edsbuf))
             && gorite()) {
          }
          else if (flgset&MULTIEX) {
               rc=0;
          }
          break;
     case HOME:
          edtxof=edtbof=0;
          break;
     case END:
          if ((strpl=strpln(edsbuf)) == mxlnm1) {
               edtbof=mxlnm1-edtedw;
               edtxof=edtedw-1;
          }
          else if (strpl >= edtbof+edtedw) {
               edtbof=strpl-edtedw+1;
               edtxof=edtedw-1;
          }
          else if (strpl < edtbof) {
               edtbof=strpl;
               edtxof=0;
          }
          else {
               edtxof=strpl-edtbof;
          }
          break;
     case INS:
          if (!(flgset&MCHOICE)) {
               cursiz((inson^=1) ? BIGCURS : LILCURS);
          }
          break;
     default:
          savexo=pxoff;
          saveyo=pyoff;
          if (inson) {
               if ((*validrou)(edtvalc,edsbuf) && strpln(edsbuf) < mxlnm1) {
                    movmem(edsbuf+edtbof+edtxof,edsbuf+edtbof+edtxof+1,
                           strpln(edsbuf+edtbof+edtxof));
                    edsbuf[edtbof+edtxof]=(flgset&ALLCAPS ? toupper(edtvalc)
                                                          : edtvalc);
                    gorite();
               }
          }
          else {
               if ((*validrou)(edtvalc,edsbuf)) {
                    edsbuf[edtbof+edtxof]=(flgset&ALLCAPS ? toupper(edtvalc)
                                                          : edtvalc);
                    gorite();
               }
          }
          pxoff=savexo;
          pyoff=saveyo;
          if (edtvalc == '\r' || ((flgset&MULTIEX) && (edtvalc == CRSRDN
                                                    || edtvalc == CRSRUP
                                                    || edtvalc == BAKTAB
                                                    || edtvalc == TAB))) {
               rc=0;
          }
     }
     return(rc);
}

STATIC int                              /*   1: could move  0: couldn't    */
gorite(void)                            /* advance cursor to right 1 space */
{
     int rc=1;

     if (edtxof < edtedw-1) {
          edtxof++;
     }
     else if (edtbof+edtedw < mxlnm1) {
          edtbof++;
     }
     else {
          rc=0;
     }
     return(rc);
}

void
endedt(void)                            /* wrap up editing session         */
{
     edtbof=0;
     rfredt(savatr);
     depad(edsbuf);
     cursiz(savecur);
}
